# Copyright 2018-2020 Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..commons import Glossary

def __init__():
    """
    Setup the module silently
    """
    from .receptors import ReceptorQuery
    from .genes import AllenBrainAtlasQuery
    from .connectivity import ConnectivityProfileExtractor, ConnectivityMatrixExtractor
    from .ebrainsquery import EbrainsRegionalFeatureExtractor
    from .extractor import FeatureExtractorRegistry
    extractor_types = FeatureExtractorRegistry() 
    return [ extractor_types,
            Glossary(AllenBrainAtlasQuery.GENE_NAMES.keys()),
            Glossary(extractor_types.modalities.keys()) ]

extractor_types,gene_names,modalities = __init__()
classes = { name:extractor_types._extractors[name][0]._FEATURETYPE 
        for name in extractor_types._extractors.keys()}



# Copyright 2018-2020 Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nibabel.affines import apply_affine
from numpy import linalg as npl
import numpy as np

from . import parcellations, spaces, features, logger
from .region import Region
from .features.feature import GlobalFeature
from .features import classes as feature_classes
from .commons import create_key,MapType
from .config import ConfigurationRegistry
from .space import Space

VERSION_BLACKLIST_WORDS=["beta","rc","alpha"]

class Atlas:

    def __init__(self,identifier,name):
        # Setup an empty. Use _add_space and _add_parcellation to complete
        # the setup.
        self.name = name
        self.id = identifier
        self.key = create_key(name)

        # no parcellation initialized at construction
        self.parcellations = [] # add with _add_parcellation
        self.spaces = [] # add with _add_space

        # nothing selected yet at construction 
        self.selected_region = None
        self.selected_parcellation = None 

    def __hash__(self):
        """
        Used for caching functions taking atlas object as an input, like FeatureExtractor.pick_selection()
        """
        hash(self.id)+hash(self.selected_parcellation.id)+hash(self.selected_region)

    def _add_space(self, space):
        self.spaces.append(space)

    def _add_parcellation(self, parcellation, select=False):
        self.parcellations.append(parcellation)
        if self.selected_parcellation is None or select:

            self.select_parcellation(parcellation)

    def __str__(self):
        return self.name

    @staticmethod
    def from_json(obj):
        """
        Provides an object hook for the json library to construct an Atlas
        object from a json stream.
        """
        if all([ 
            '@id' in obj, 
            'spaces' in obj, 
            'parcellations' in obj ]):
            p = Atlas(obj['@id'], obj['name'])
            for space_id in obj['spaces']:
                if space_id not in spaces:
                    raise ValueError(f"Invalid atlas configuration for {str(p)} - space {space_id} not known")
                p._add_space( spaces[space_id] )
            for parcellation_id in obj['parcellations']:
                if parcellation_id not in parcellations:
                    raise ValueError(f"Invalid atlas configuration for {str(p)} - parcellation {parcellation_id} not known")
                p._add_parcellation( parcellations[parcellation_id] )
            return p
        return obj

    @property
    def regionnames(self):
        return self.selected_parcellation.names

    @property
    def regionlabels(self):
        return self.selected_parcellation.regiontree.labels


    def threshold_continuous_maps(self,threshold):
        """
        Inform the atlas that thresholded continuous maps should be preferred
        over static labelled maps for building and using region masks.
        This will, for example, influence spatial filtering of coordinate-based
        features in the get_features() method.
        """
        self.selected_parcellation.continuous_map_threshold = threshold

    def select(self,parcellation=None, region=None, force=False):
        """
        Select a parcellation and/or region. See Atlas.select_parcellation, Atlas.select_region
        """
        if parcellation is not None:
            self.select_parcellation(parcellation,force)
        if region is not None:
            self.select_region(region)
            
    def select_parcellation(self, parcellation,force=False):
        """
        Select a different parcellation for the atlas.

        Parameters
        ----------

        parcellation : Parcellation
            The new parcellation to be selected
        force : boolean
            If a parcellation is labelled "beta","rc" or "alpha", it cannot be selected unless force=True is provided as a positional
            argument.

        Yields
        ------
        Selected parcellation
        """
        parcellation_obj = parcellations[parcellation]
        if parcellation_obj.version is not None:
            versionname = parcellation_obj.version.name
            if any(w in versionname for w in VERSION_BLACKLIST_WORDS) and not force:
                logger.warning(f"Will not select experimental version {versionname} of {parcellation_obj.name} unless forced.")
                return
        if parcellation_obj not in self.parcellations:
            logger.error('The requested parcellation is not supported by the selected atlas.')
            logger.error('    Parcellation:  '+parcellation_obj.name)
            logger.error('    Atlas:         '+self.name)
            logger.error(parcellation_obj.id,self.parcellations)
            raise Exception('Invalid Parcellation')
        self.selected_parcellation = parcellation_obj
        self.selected_region = parcellation_obj.regiontree
        logger.info(f'Select "{self.selected_parcellation}"')
        return self.selected_parcellation

    def select_region(self,region):
        """
        Selects a particular region. 

        TODO test carefully for selections of branching points in the region
        hierarchy, then managing all regions under the tree. This is nontrivial
        because for incomplete parcellations, the union of all child regions
        might not represent the complete parent node in the hierarchy.

        Parameters
        ----------
        region : Region
            Region to be selected. Both a region object, as well as a region
            key (uppercase string identifier) are accepted.

        Yields
        ------
        Selected region
        """
        previous_selection = self.selected_region
        if isinstance(region,Region):
            # argument is already a region object - use it
            self.selected_region = region
        else:
            # try to interpret argument as the key for a region 
            selected = self.selected_parcellation.regiontree.find(
                    region,select_uppermost=True)
            if len(selected)==1:
                # one match found - fine
                self.selected_region = next(iter(selected))
            elif len(selected)==0:
                # no match found
                logger.error('Cannot select region. The spec "{}" does not match any known region.'.format(region))
            else:
                # multiple matches found. We do not allow this for now.
                logger.error('Cannot select region. The spec "{}" is not unique. It matches: {}'.format(
                    region,", ".join([s.name for s in selected])))
        if not self.selected_region == previous_selection:
            logger.info(f'Select "{self.selected_region.name}"')
        return self.selected_region

    def get_map(self, space=None, maptype=MapType.LABELLED):
        """
        return the map provided by the selected parcellation in the given space.
        This just forwards to the selected parcellation object, see
        Parcellation.get_map()
        """
        return self.selected_parcellation.get_map(space=space,maptype=maptype)

    def build_mask(self, space : Space, resolution_mm=None ):
        """
        Returns a binary mask in the given space, where nonzero values denote
        voxels corresponding to the current region selection of the atlas. 

        WARNING: Note that for selections of subtrees of the region hierarchy, this
        might include holes if the leaf regions are not completly covering
        their parent and the parent itself has no label index in the map.

        Parameters
        ----------
        space : Space
            Template space 
        resolution_mm : float or None (Default: None)
            Request the template at a particular physical resolution. If None,
            the native resolution is used.
            Currently, this only works for the BigBrain volume.
        """
        return self.selected_region.build_mask(space,resolution_mm=resolution_mm)

    def get_template(self,space=None):
        """
        Get the volumetric reference template image for the given space.

        See
        ---
        Space.get_template()

        Parameters
        ----------
        space : str
            Template space definition, given as a dictionary with an '@id' key

        Yields
        ------
        A nibabel Nifti object representing the reference template, or None if not available.
        TODO Returning None is not ideal, requires to implement a test on the other side. 
        """
        if space is None:
            space = self.spaces[0]
            if len(self.spaces)>1:
                logger.warning(f'{self.name} supports multiple spaces, but none was specified. Falling back to {space.name}.')


        try:
            spaceobj = spaces[space]
        except IndexError:
            logger.error(f'Atlas "{self.name}" does not support reference space "{space}".')
            print("Available spaces:")
            for space in self.spaces:
                print(space.name,space.id)
            return None

        return spaceobj.get_template()

    def decode_region(self,regionspec):
        """
        Given a unique specification, return the corresponding region from the selected parcellation.
        The spec could be a label index, a (possibly incomplete) name, or a
        region object.
        This method is meant to definitely determine a valid region. Therefore, 
        if no match is found, it raises a ValueError. If it finds multiple
        matches, it tries to return only the common parent node. If there are
        multiple remaining parent nodes, which is rare, a custom group region is constructed.

        Parameters
        ----------
        regionspec : any of 
            - a string with a possibly inexact name, which is matched both
              against the name and the identifier key, 
            - an integer, which is interpreted as a labelindex,
            - a region object

        Return
        ------
        Region object
        """
        return self.selected_parcellation.decode_region(regionspec)

    def find_regions(self,regionspec,all_parcellations=False):
        """
        Find regions with the given specification in the current or all
        available parcellations of this atlas.

        Parameters
        ----------
        regionspec : any of 
            - a string with a possibly inexact name, which is matched both
              against the name and the identifier key, 
            - an integer, which is interpreted as a labelindex
            - a region object
        all_parcellations : Boolean (default:False)
            Do not only search the selected but instead all available parcellations.

        Yield
        -----
        list of matching regions
        """
        if not all_parcellations:
            return self.selected_parcellation.find_regions(regionspec)
        result = []
        for p in self.parcellations:
            result.extend(p.find_regions(regionspec))
        return result

    def clear_selection(self):
        """
        Cancels any current region selection.
        """
        self.select_region(self.selected_parcellation.regiontree)

    def region_selected(self,region):
        """
        Verifies wether a given region is part of the current selection.
        """
        return self.selected_region.includes(region)

    def coordinate_selected(self,space,coordinate):
        """
        Verifies wether a position in the given space is part of the current
        selection.

        Parameters
        ----------
        space : Space
            The template space in which the test shall be carried out
        coordinate : tuple x/y/z
            A coordinate position given in the physical space. It will be
            converted to the voxel space using the inverse affine matrix of the
            template space for the query.

        NOTE: since get_mask is lru-cached, this is not necessary slow
        """
        assert(space in self.spaces)
        # transform physical coordinates to voxel coordinates for the query
        def check(mask):
            voxel = (apply_affine(npl.inv(mask.affine),coordinate)+.5).astype(int)
            if np.any(voxel>=mask.dataobj.shape):
                return False
            if mask.dataobj[voxel[0],voxel[1],voxel[2]]==0:
                return False
            return True
        M = self.build_mask(space)
        return any(check(M.slicer[:,:,:,i]) for i in range(M.shape[3])) if M.ndim==4 else check(M)

    def get_features(self,modality,**kwargs):
        """
        Retrieve data features linked to the selected atlas configuration, by modality. 
        See siibra.features.modalities for available modalities.
        """
        hits = []

        if modality not in features.extractor_types.modalities:
            logger.error("Cannot query features - no feature extractor known "\
                    "for feature type {}.".format(modality))
            return hits

        for cls in features.extractor_types[modality]:
            if modality=='GeneExpression':
                extractor = cls(self,kwargs['gene'])
            else:
                extractor = cls(self)
            hits.extend(extractor.pick_selection(self))

        return list(set(hits))

    def assign_coordinates(self,space:Space,xyz_mm,sigma_mm=3):
        """
        Assign physical coordinates with optional standard deviation to atlas regions.
        See also: ContinuousParcellationMap.assign_coordinates()

        Parameters
        ----------
        space : Space
            reference template space for computing the assignemnt
        xyz_mm : coordinate tuple 
            3D point in physical coordinates of the template space of the
            ParcellationMap. Also accepts a string of the format "15.453mm, 4.828mm, 69.122mm" 
            as copied from siibra-explorer.
        sigma_mm : float (default: 0)
            standard deviation /expected localization accuracy of the point, in
            physical units. If nonzero, A 3D Gaussian distribution with that
            bandwidth will be used for representing the location instead of a
            deterministic coordinate.
        """
        smap = self.selected_parcellation.get_map(space,maptype=MapType.CONTINUOUS)
        return smap.assign_coordinates(xyz_mm, sigma_mm)

    def assign_maps(self,space:Space,mapimg):
        """
        Assign physical coordinates with optional standard deviation to atlas regions.
        See also: ContinuousParcellationMap.assign_coordinates()

        Parameters
        ----------
        space : Space
            reference template space for computing the assignemnt
        mapimg : 3D volume as nibabel spatial image
        """
        smap = self.selected_parcellation.get_map(space,maptype=MapType.CONTINUOUS)
        return smap.assign(mapimg)


REGISTRY = ConfigurationRegistry('atlases', Atlas)

if __name__ == '__main__':

    atlas = REGISTRY.MULTILEVEL_HUMAN_ATLAS




# Copyright 2018-2020 Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .commons import create_key,HasOriginDataInfo,OriginDataInfo
from .config import ConfigurationRegistry
import numpy as np
from . import volumesrc
from . import logger
from . import arrays
import copy
from cloudvolume import Bbox
from typing import Tuple
import nibabel as nib

class Space(HasOriginDataInfo):
    """
    A particular brain reference space.
    """

    def __init__(self, identifier, name, template_type=None, src_volume_type=None, volume_src={}):
        HasOriginDataInfo.__init__(self)
        self.id = identifier
        self._rename(name)
        self.type = template_type
        self.src_volume_type = src_volume_type
        self.volume_src = volume_src
        self._assign_volume_sources(volume_src)

    def _assign_volume_sources(self,volume_src):
        self.volume_src = copy.deepcopy(volume_src)
        for volsrc in self.volume_src:
            try:
                volsrc.space = self
            except AttributeError as e:
                print(volsrc)
                raise(e)

    def _rename(self,newname):
        self.name = newname
        self.key = create_key(self.name)

    def __str__(self):
        return self.name

    def __repr__(self):
        return str(self)

    def get_template(self, resolution_mm=None ):
        """
        Get the volumetric reference template image for this space.

        Parameters
        ----------
        resolution_mm : float or None (Default: None)
            Request the template at a particular physical resolution in mm. If None,
            the native resolution is used.
            Currently, this only works for the BigBrain volume.

        Yields
        ------
        A nibabel Nifti object representing the reference template, or None if not available.
        TODO Returning None is not ideal, requires to implement a test on the other side. 
        """
        candidates = [vsrc for vsrc in self.volume_src if vsrc.volume_type==self.type]
        if not len(candidates)==1:
            raise RuntimeError(f"Could not resolve template image for {self.name}. This is most probably due to a misconfiguration of the volume src.")
        return candidates[0]

    def __getitem__(self,slices:Tuple[slice,slice,slice]):
        """
        Get a volume of interest specification from this space.

        Arguments
        ---------
        slices: triple of slice
            defines the x, y and z range
        """
        if len(slices)!=3:
            raise TypeError("Slice access to spaces needs to define x,y and z ranges (e.g. Space[10:30,0:10,200:300])")
        return SpaceVOI(self,[s.start for s in slices],[s.stop for s in slices])

    def get_voi(self,minpt:Tuple[float,float,float],maxpt:Tuple[float,float,float]):
        """
        Get a rectangular volume of interest specification for this space.

        Arguments
        ---------

        minpt: 3-tuple
            smaller 3D point defining the VOI
        maxpt: 3-tuple
            larger 3D point defining the VOI
        """
        return self[
            minpt[0]:maxpt[0],
            minpt[1]:maxpt[1],
            minpt[2]:maxpt[2]]

    @staticmethod
    def from_json(obj):
        """
        Provides an object hook for the json library to construct a Space
        object from a json stream.
        """
        required_keys = ['@id','name','shortName','templateType']
        if any([k not in obj for k in required_keys]):
            return obj

        if "minds/core/referencespace/v1.0.0" not in obj['@id']:
            return obj

        volume_src = [volumesrc.from_json(v) for v in obj['volumeSrc']] if 'volumeSrc' in obj else []
        s=Space(obj['@id'], obj['shortName'], template_type = obj['templateType'],
                src_volume_type = obj.get('srcVolumeType'),
                volume_src = volume_src)
        origin_datainfos=[OriginDataInfo.from_json(f) for f in obj.get('originDatasets', [])]
        s.origin_datainfos=[f for f in origin_datainfos if f is not None]
        return s


class SpaceVOI(Bbox):

    def __init__(self,space:Space,minpt:Tuple[float,float,float],maxpt:Tuple[float,float,float]):
        super().__init__(minpt,maxpt)
        self.space = space

    @staticmethod
    def from_map(space:Space,roi:nib.Nifti1Image):
        # construct from a roi mask or map
        bbox = arrays.bbox3d(roi.dataobj,affine=roi.affine)
        return SpaceVOI(space,bbox[:3,0],bbox[:3,1])

    def transform_bbox(self,transform):
        assert(transform.shape==(4,4))
        return Bbox(
            np.dot(transform,np.r_[self.minpt,1])[:3].astype('int'),
            np.dot(transform,np.r_[self.maxpt,1])[:3].astype('int') )

    def __str__(self):
        return f"Bounding box {self.minpt}mm -> {self.maxpt}mm defined in {self.space.name}"

    def __repr__(self):
        return str(self)


REGISTRY = ConfigurationRegistry('spaces', Space)



# Copyright 2018-2020 Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# TODO this file lacks documentation

import json
from .feature import RegionalFeature
from .extractor import FeatureExtractor
import re
import os
from .. import ebrains
from .. import logger

IGNORE_PROJECTS = [
  'Julich-Brain: cytoarchitectonic probabilistic maps of the human brain'
]

kg_feature_summary_kwargs={
    'org': 'minds',
    'domain': 'core',
    'schema': 'parcellationregion',
    'version': 'v1.0.0'
}

KG_REGIONAL_FEATURE_SUMMARY_QUERY_NAME = 'siibra-kg-feature-summary-0.0.1'

def get_dataset(parcellation=None):
    if parcellation is None:
        raise ValueError('parcellation is required to find_regions')

    result=ebrains.execute_query_by_id(query_id=KG_REGIONAL_FEATURE_SUMMARY_QUERY_NAME,
        **ebrains.kg_feature_query_kwargs,**kg_feature_summary_kwargs)

    return_list=[]
    for r in result.get('results', []):
        for dataset in r.get('datasets', []):
            ds_id = dataset.get('@id')
            ds_name = dataset.get('name')
            ds_embargo_status=dataset.get('embargo_status')
            if not "dataset" in ds_id:
                logger.debug(f"'{ds_name}' is not an interpretable dataset and will be skipped.\n(id:{ds_id})")
                continue
            regionname = r.get('name', None)
            regions = parcellation.find_regions(regionname)
            for region in regions:
                return_list.append((region, ds_id, ds_name, ds_embargo_status))
    return return_list

class EbrainsRegionalDataset(RegionalFeature, ebrains.EbrainsDataset):

    @staticmethod
    def preheat(id=None):
        from .. import parcellations
        if id is not None and parcellations[id] is not None:
            get_dataset(parcellations[id])
        else:
            for p in parcellations:
                get_dataset(p)

    def __init__(self, region, id, name, embargo_status):
        RegionalFeature.__init__(self, region)
        ebrains.EbrainsDataset.__init__(self, id, name, embargo_status)

    def __str__(self):
        return ebrains.EbrainsDataset.__str__(self)

    def __hash__(self):
        return ebrains.EbrainsDataset.__hash__(self)

class EbrainsRegionalFeatureExtractor(FeatureExtractor):
    _FEATURETYPE=EbrainsRegionalDataset
    def __init__(self, atlas):
        FeatureExtractor.__init__(self,atlas)
        
        # potentially, using ebrains_id is a lot quicker
        # but if user selects region with higher level of hierarchy, this benefit may be offset by numerous http calls
        # even if they were cached...
        # ebrains_id=atlas.selected_region.attrs.get('fullId', {}).get('kg', {}).get('kgId', None)
        for region, ds_id, ds_name, ds_embargo_status in get_dataset(atlas.selected_parcellation):
            feature = EbrainsRegionalDataset(region=region, id=ds_id, name=ds_name,
                embargo_status=ds_embargo_status)
            self.register(feature)


def set_specs():
    MODULE_DIR=os.path.dirname(os.path.abspath(__file__))
    with open(os.path.join(MODULE_DIR,'ebrainsquery_specs.json'),'r') as f:
        QUERYSPEC=json.load(f)
    req = ebrains.upload_schema(
        query_id=KG_REGIONAL_FEATURE_SUMMARY_QUERY_NAME,
        spec=QUERYSPEC['summary_spec'],
        **kg_feature_summary_kwargs)
    if not req.status_code < 400:
        raise RuntimeError("Could not upload query")

if __name__ == '__main__':
    pass



# Copyright 2018-2020 Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from os import path, getenv

# __version__ is parsed by setup.py
__version__='0.1a7'

import logging
logger = logging.getLogger(__name__)
ch = logging.StreamHandler()
formatter = logging.Formatter('[%(name)s:%(levelname)s]  %(message)s')
ch.setFormatter(formatter)
logger.addHandler(ch)

class LoggingContext:
    def __init__(self, level ):
        self.level = level

    def __enter__(self):
        self.old_level = logger.level
        logger.setLevel(self.level)

    def __exit__(self, et, ev, tb):
        logger.setLevel(self.old_level)

QUIET = LoggingContext("ERROR")
VERBOSE = LoggingContext("DEBUG")

# convenience function, will be easier to discover
def set_log_level(level):
    logger.setLevel(level)

set_log_level(getenv("SIIBRA_LOG_LEVEL","INFO"))

logger.info(f"Version: {__version__}")
logger.warning("This is a development release. Use at your own risk. Please file bugs and issues at https://github.com/FZJ-INM1-BDA/siibra-python.")

from .space import REGISTRY as spaces
from .parcellation import REGISTRY as parcellations
from .atlas import REGISTRY as atlases
from .retrieval import clear_cache
from .features import modalities
from .ebrains import set_token as set_ebrains_token
from .commons import MapType,ParcellationIndex


import os
import unittest
import siibra as sb
from test.get_token import get_token

token = get_token()
os.environ['HBP_AUTH_TOKEN'] = token["access_token"]

class TestRetrievalDownloadFile(unittest.TestCase):

    def test_download_file(self):
        sb.retrieval.download_file(
            "https://object.cscs.ch/v1/AUTH_227176556f3c4bb38df9feea4b91200c/test_stefan_destination/MPM.zip",
            ziptarget="JulichBrain_v25.xml"
        )
    # TODO Clear cache folder after test (for local testing)

if __name__ == "__main__":
    unittest.main()



import unittest
import siibra

class TestParcellationVersion(unittest.TestCase):
    correct_json={
        'name': 'foobar',
        'collectionName': 'foobar-collection',
        '@prev': 'foobar-prev',
        '@next': 'foobar-next',
        'deprecated': False,
    }
    def test_from_json(self):

        ver=siibra.parcellation.ParcellationVersion.from_json(self.correct_json)
        self.assertTrue(ver.deprecated == self.correct_json['deprecated'])
        self.assertTrue(ver.name == self.correct_json['name'])
        self.assertTrue(ver.collection == self.correct_json['collectionName'])
        self.assertTrue(ver.prev_id == self.correct_json['@prev'])
        self.assertTrue(ver.next_id == self.correct_json['@next'])
        # TODO test prev/next
        


if __name__ == "__main__":
    unittest.main()



# Copyright 2018-2020 Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from zipfile import ZipFile
import requests
import hashlib
import os
#from os import os.path
from . import logger

# TODO unify download_file and cached_get
# TODO manage the cache: limit total memory used, remove old zips,...

# Ideas:
#
# this module can detect several flavors of download file specifactions, for example:
# - A nifti file on a standard http URL, no auth required -> just download it
# - A zip file on a standard http URL -> We need additional specificcation of the desired file in side the zip
#   (needs to be reflected in the metadata scheme for e.g. spaces and parcellations)
# - a UID of data provider like EBRAINS -> need to detecto the provider, and use an auth key
#   that is setup during package installation
#


def __compile_cachedir(suffix=None):
    from os import path,makedirs,environ
    from appdirs import user_cache_dir
    if "SIIBRA_CACHEDIR" in environ:
        cachedir = environ['SIIBRA_CACHEDIR']
    else:
        dname = __name__ if suffix is None else __name__+"_"+suffix
        cachedir = user_cache_dir(__name__,"")
    # make sure cachedir exists and is writable
    try:
        if not path.isdir(cachedir):
            makedirs(cachedir)
        assert(os.access(cachedir, os.W_OK))
    except Exception as e:
        print(str(e))
        raise PermissionError('Cannot create local cache folder at {}. Please set a writable cache directory using the environment variable SIIBRA_CACHEDIR, and reload siibra.'.format(cachedir))
    return cachedir

CACHEDIR = __compile_cachedir()
logger.debug('Using cache: {}'.format(CACHEDIR))

hashstr = lambda s: str(hashlib.sha256(s).hexdigest())

def cachefile(str_rep,suffix=None):
    hashfile = os.path.join(CACHEDIR,hashstr(str_rep))
    if suffix is None:
        return hashfile
    else:
        return hashfile+"."+suffix


def download_file(url, ziptarget=None, targetname=None ):
    """
    Downloads a file from a URL to local disk, and returns the local filename.

    Parameters
    ----------
    url : string
        Download link
    ziptarget : string
        [Optional] If the download gives a zip archive, this paramter gives the
        filename to be extracted from the archive.
    targetname : string (optional)
        Desired filename after download.

    TODO Handle and write tests for non-existing URLs, password-protected URLs, too large files, etc.
    """

    # Existing downloads are indicated by a hashfile generated from the URL,
    # which includes the filename of the actual image. This is a workaround to
    # deal with the fact that we do not know the filetype prior to downloading,
    # so we cannot determine the suffix in advance.
    url_str = url if ziptarget is None else url+ziptarget
    hashfile = cachefile(str.encode(url_str))
    if os.path.exists(hashfile):
        with open(hashfile, 'r') as f:
            filename,cachename = f.read().split(';')
            if os.path.exists(cachename):
                return cachename

    # No valid hash and corresponding file found - need to download
    req = requests.get(url)
    if req is not None and req.status_code == 200:

        # try to figure out the original filename for the requested file
        if targetname is not None:
            original_filename = targetname
        elif 'X-Object-Meta-Orig-Filename' in req.headers:
            original_filename = req.headers['X-Object-Meta-Orig-Filename']
        else:
            original_filename = os.path.basename(url)

        # build a uid-based alternative filename for the cache which is not redundant
        # (but keep the original filename for reference)
        namefields = os.path.basename(original_filename).split(".")
        suffix =  ".".join(namefields[1:]) if len(namefields)>1 else ".".join(namefields)
        cachename = hashfile+"."+suffix
        filename = original_filename

        # now save the file
        with open(cachename, 'wb') as code:
            code.write(req.content)

        # if this was a zip file, and a particular target file in the zip was
        # requested, we need to extract it now. We will later drop the zipfile.
        if suffix.endswith("zip") and (ziptarget is not None):
            filename = ziptarget
            cachename = get_from_zip(
                    cachename, ziptarget)
        with open(hashfile, 'w') as f:
            f.write(filename+";")
            f.write(cachename)
        return cachename
    '''
        - error on response status != 200
        - error on file read
        - Nibable error
        - handle error, when no filename header is set
        - error or None when space not known
        - unexpected error
        '''


def get_from_zip(zipfile, ziptarget ):
    # Extract temporary zip file
    # TODO catch problem if file is not a nifti
    targetname = None
    with ZipFile(zipfile, 'r') as zip_ref:
        for zip_info in zip_ref.infolist():
            if zip_info.filename[-1] == '/':
                continue
            zip_info.filename = os.path.basename(zip_info.filename)
            if zip_info.filename == ziptarget:
                downloadname = CACHEDIR + '/' + zip_info.filename
                zip_ref.extract(zip_info, CACHEDIR)
                targetname = "{}.{}".format(
                    os.path.splitext(zipfile)[0],
                    os.path.basename(ziptarget) )
                if not os.path.exists(targetname):
                    os.rename(downloadname,targetname)
    os.remove(zipfile)
    if targetname is not None:
        return targetname
    raise Exception("target file",ziptarget,"not found in zip archive",zipfile)


def get_json_from_url(url):
    req = requests.get(url)
    if req is not None and req.status_code == 200:
        return json.loads(req.content)
    else:
        return {}


def cached_get(url,msg_if_not_cached=None,**kwargs):
    """
    Performs a requests.get if the result is not yet available in the local
    cache, otherwise returns the result from the cache.
    This leaves the interpretation of the returned content to the caller.
    TODO we might extend this as a general tool for the siibra library, and make it a decorator
    TODO this hashes the authentication token as part of kwargs, which might not be desired
    FIXME The error messages are specific to EBRAINS queries, but this method is not.
    """
    logger.debug(f"Get from URL\n{url}")
    logger.debug(f"kwargs: {kwargs}")
    url_hash = hashlib.sha256((url+json.dumps(kwargs)).encode('ascii')).hexdigest()
    cachefile_content = os.path.join(CACHEDIR,url_hash)+".content"
    cachefile_url = os.path.join(CACHEDIR,url_hash)+".url"

    if os.path.isfile(cachefile_content):
        # This URL target is already in the cache - just return it
        logger.debug("Returning cached response of url {} at {}".format(url,cachefile_content))
        with open(cachefile_content,'rb') as f:
            r = f.read()
            return(r)
    else:
        if msg_if_not_cached:
            print(msg_if_not_cached)
        r = requests.get(url,**kwargs)
        if r.ok:
            with open(cachefile_content,'wb') as f:
                f.write(r.content)
            with open(cachefile_url,'w') as f:
                f.write(url)
            return r.content
        elif r.status_code == 401:
            logger.error('The provided authentication token is not valid')
        elif r.status_code == 403:
            logger.error('No permission to access the given query')
        elif r.status_code == 404:
            logger.error('Query with this id not found')
        else:
            logger.error('Problem with "get" protocol on url: %s ' % url )
        raise RuntimeError('Could not retrieve data.')

def clear_cache():
    import shutil
    logger.info("Clearing siibra cache.")
    shutil.rmtree(CACHEDIR)
    __compile_cachedir()



# Copyright 2018-2020 Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import numbers
from scipy.ndimage import gaussian_filter
import nibabel as nib
import re
from memoization import cached
from hashlib import sha1

def __bbox(A):
    """
    Bounding box of nonzero values in a 3D array.
    https://stackoverflow.com/questions/31400769/bounding-box-of-numpy-array

    If affine not None, the affine is applied to the bounding box.
    """
    x = np.any(A, axis=(1, 2))
    y = np.any(A, axis=(0, 2))
    z = np.any(A, axis=(0, 1))
    nzx,nzy,nzz = [np.where(v) for v in (x,y,z)]
    if any(len(nz[0])==0 for nz in [nzx,nzy,nzz]):
        # empty array
        return None
    xmin, xmax = nzx[0][[0, -1]]
    ymin, ymax = nzy[0][[0, -1]]
    zmin, zmax = nzz[0][[0, -1]]
    return np.array([
        [xmin, xmax+1], 
        [ymin, ymax+1], 
        [zmin, zmax+1],
        [1,1] ])

def bbox3d(A,affine=np.identity(4)):
    bbox = np.dot(affine,__bbox(A))
    return np.sort(bbox.astype('int'))

def bbox_intersection(bb1,bb2):
    lower = [max(bb1[i,0],bb2[i,0]) for i in range(3)]
    upper = np.maximum(lower,[min(bb1[i,1],bb2[i,1]) for i in range(3)])
    return np.prod(upper-lower)

def parse_coordinate_str(cstr:str,unit='mm'):
    pat=r'([-\d\.]*)'+unit
    digits = re.findall(pat,cstr)
    if len(digits)==3:
        return np.array([float(d) for d in digits])
    else:
        return None

def create_homogeneous_array(xyz):
    """
    From one 3D point or list of multiple 3D points, build the Nx4 homogenous coordinate array.

    Parameters
    ----------

    xyz_phys : 3D coordinate tuple, list of 3D tuples, or Nx3 array of coordinate tuples
        3D point(s) in physical coordinates of the template space of the
        ParcellationMap
    """
    if isinstance(xyz,str):
        parsed = parse_coordinate_str(xyz)
        if parsed is not None:
            XYZH = np.ones((1,4))
            XYZH[0,:3] = parsed
            return XYZH
    elif isinstance(xyz,list) and all(isinstance(e,str) for e in xyz):
        parsed = (parse_coordinate_str(s) for s in xyz)
        valid = [np.r_[p,1] for p in parsed if p is not None]
        if len(valid)>0:
            return np.array(valid)
    elif isinstance(xyz[0],numbers.Number):
        # only a single point provided
        assert(len(xyz) in [3,4])
        XYZH = np.ones((1,4))
        XYZH[0,:len(xyz)] = xyz
        return XYZH
    else:
        # assume list of coordinates
        XYZ = np.array(xyz)
        assert(XYZ.shape[1]==3)
        return np.c_[XYZ,np.ones_like(XYZ[:,0])]
    raise ValueError(f"Could not parse xyz coordinates: {xyz}")

def assert_homogeneous_3d(xyz):
    if len(xyz)==4:
        return xyz
    else:
        return np.r_[xyz,1]

def create_gaussian_kernel(sigma=1,sigma_point=3):
    """
    Compute a 3D Gaussian kernel of the given bandwidth.
    """
    r = int(sigma_point*sigma)
    k_size = 2*r + 1
    impulse = np.zeros((k_size,k_size,k_size))
    impulse[r,r,r] = 1
    kernel = gaussian_filter(impulse, sigma)
    kernel /= kernel.sum()
    return kernel

def argmax_dim4(img,dim=-1):
    """
    Given a nifti image object with four dimensions, returns a modified object
    with 3 dimensions that is obtained by taking the argmax along one of the
    four dimensions (default: the last one). To distinguish the pure background
    voxels from the foreground voxels of channel 0, the argmax indices are
    incremented by 1 and label index 0 is kept to represent the background.
    """
    assert(len(img.shape)==4)
    assert(dim>=-1 and dim<4)
    newarr = np.asarray(img.dataobj).argmax(dim)+1
    # reset the true background voxels to zero
    newarr[np.asarray(img.dataobj).max(dim)==0]=0
    return nib.Nifti1Image(
            dataobj = newarr,
            header = img.header,
            affine = img.affine )


def MI(arr1,arr2,nbins=100,normalized=True):
    """
    Compute the mutual information between two 3D arrays, which need to have the same shape.

    Parameters:
    arr1 : First 3D array
    arr2 : Second 3D array
    nbins : number of bins to use for computing the joint histogram (applies to intensity range)
    normalized : Boolean, default:True
        if True, the normalized MI of arrays X and Y will be returned, 
        leading to a range of values between 0 and 1. Normalization is 
        achieved by NMI = 2*MI(X,Y) / (H(X) + H(Y)), where  H(x) is the entropy of X
    """

    assert(all(len(arr.shape)==3 for arr in [arr1,arr2]))
    assert(all(arr.size>0) for arr in [arr1,arr2])

    # compute the normalized joint 2D histogram as an 
    # empirical measure of the joint probabily of arr1 and arr2
    pxy, _, _ = np.histogram2d(arr1.ravel(),arr2.ravel(),bins=nbins)
    pxy /= pxy.sum()
    
    # extract the empirical propabilities of intensities 
    # from the joint histogram
    px = np.sum(pxy, axis=1) # marginal for x over y
    py = np.sum(pxy, axis=0) # marginal for y over x
    
    # compute the mutual information
    px_py = px[:, None] * py[None, :] 
    nzs = pxy > 0 # nonzero value indices
    I = np.sum(pxy[nzs] * np.log(pxy[nzs] / px_py[nzs]))
    if not normalized:
        return I
    
    # normalize, using the sum of their individual entropies H
    def entropy(p):
        nz = p>0
        assert(np.count_nonzero(nz)>0)
        return -np.sum(p[nz]*np.log(p[nz]))
    Hx,Hy = [entropy(p) for p in [px,py]]
    assert((Hx+Hy)>0)
    NMI = 2*I / (Hx+Hy)
    return NMI


# Copyright 2018-2020 Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC,abstractmethod
from .. import logger
from ..region import Region

class Feature(ABC):
    """ 
    Abstract base class for all data features.
    """

    @abstractmethod
    def matches(self,atlas):
        """
        Returns True if this feature should be considered part of the current
        selection of the atlas object, otherwise else.
        """
        pass

    @abstractmethod
    def __str__(self):
        """
        Print a reasonable name of this feature.
        """
        return 

    def __hash__(self):
        return hash(str(self))

class SpatialFeature(Feature):
    """
    Base class for coordinate-anchored data features.
    """

    def __init__(self,space,location):
        self.space = space
        self.location = location
 
    def matches(self,atlas):
        """
        Returns true if the location of this feature is inside the selected
        region of the atlas, according to the mask in the reference space.
        """
        return atlas.coordinate_selected(self.space,self.location)

    def __str__(self):
        return "Features in '{space}' at {loc[0]}/{loc[1]}/{loc[2]}".format(
                space=self.space, loc=self.location)

class RegionalFeature(Feature):
    """
    Base class for region-anchored data features (semantic anchoring to region
    names instead of coordinates).
    TODO store region as an object that has a link to the parcellation
    """

    def __init__(self,region : Region):
        assert(isinstance(region,Region))
        self.region = region
 
    def matches(self,atlas):
        """
        Returns true if this feature is linked to the currently selected region
        in the atlas.
        """
        for r in self.region:
            matching_regions = atlas.selected_region.find(r)
            if len(matching_regions)>0:
                return True

class GlobalFeature(Feature):
    """
    Base class for data features which apply to the atlas as a whole
    instead of a particular location or region. A typical example is a
    connectivity matrix, which applies to all regions in the atlas.
    """

    def __init__(self,parcellation):
        self.parcellation = parcellation
 
    def matches(self,atlas):
        """
        Returns true if this global feature is related to the given atlas.
        """
        if self.parcellation == atlas.selected_parcellation:
            return True




# Copyright 2018-2020 Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from enum import Enum
from abc import ABC, abstractmethod
from . import logger

class Glossary:
    """
    A very simple class that provides enum-like simple autocompletion for an
    arbitrary list of names.
    """
    def __init__(self,words):
        self.words = list(words)

    def __dir__(self):
        return self.words

    def __str__(self):
        return "\n".join(self.words)

    def __iter__(self):
        return (w for w in self.words)

    def __contains__(self,index):
        return index in self.__dir__()

    def __getattr__(self,name):
        if name in self.words:
            return name
        else:
            raise AttributeError("No such term: {}".format(name))

def create_key(name):
    """
    Creates an uppercase identifier string that includes only alphanumeric
    characters and underscore from a natural language name.
    """
    return re.sub(
            r' +','_',
            "".join([e if e.isalnum() else " " 
                for e in name]).upper().strip() 
            )

class MapType(Enum):
    LABELLED = 1
    CONTINUOUS = 2


class ParcellationIndex:
    """
    Identifies a unique region in a ParcellationMap, combining its labelindex (the "color") and mapindex (the number of the 3Dd map, in case multiple are provided).
    """
    def __init__(self,map,label):
        self.map=map
        self.label=label
    
    def __str__(self):
        return f"({self.map}/{self.label})"

    def __repr__(self):
        return f"{self.__class__.__name__} "+str(self)

    def __eq__(self,other):
        return all([self.map==other.map,self.label==other.label])

    def __hash__(self):
        return hash((self.map,self.label))

class OriginDataInfo:
    def __init__(self, name=None, description=None, urls=[]):
        self.name=name
        self.description=description
        self.urls=urls

    @staticmethod
    def from_json(jsonstr):
        json_type=jsonstr.get('@type')
        if json_type == 'fzj/tmp/simpleOriginInfo/v0.0.1':
            return OriginDataInfo(name=jsonstr.get('name'),
                        description=jsonstr.get('description'),
                        urls=jsonstr.get('url', []))
        if json_type == 'minds/core/dataset/v1.0.0':
            from .ebrains import EbrainsDatasetOriginDataInfo
            return EbrainsDatasetOriginDataInfo(id=jsonstr.get('kgId'))
        logger.debug(f'Cannot parse {jsonstr}')
        return None

class HasOriginDataInfo:
    def __init__(self):
        self.origin_datainfos = []


import unittest

import siibra as sb
from siibra.space import Space

class TestSpaces(unittest.TestCase):

    space_id = 'space_id'
    name = 'space name'
    url = 'space_url'
    ziptarget = 'space_zip_target'
    template = 'temp_file'
    ttype = 'nii'

    json_space_with_zip = {
        '@id': 'space1/minds/core/referencespace/v1.0.0',
        'name': name,
        'shortName': name,
        'templateType': ttype,
        "volumeSrc": [{
            "@type": "fzj/tmp/volume_type/v0.0.1",
            "@id": "fzj/tmp/volume_type/v0.0.1/icbm152_2009c_nonlin_asym/nifti",
            "name": "icbm152_2009c_nonlin_asym/nifti",
            "volume_type": ttype,
            "url": url,
            "zipped_file": ziptarget,
            }]
        }

    json_space_without_zip = {
        '@id': 'space1/minds/core/referencespace/v1.0.0',
        'name': name,
        'shortName': name,
        'templateUrl': url,
        'templateFile': ziptarget,
        'templateType': ttype,
        "volumeSrc": [{
            "@type": "fzj/tmp/volume_type/v0.0.1",
            "@id": "fzj/tmp/volume_type/v0.0.1/icbm152_2009c_nonlin_asym/nifti",
            "name": "icbm152_2009c_nonlin_asym/nifti",
            "volume_type": "nii",
            "url": "http://www.bic.mni.mcgill.ca/~vfonov/icbm/2009/mni_icbm152_nlin_asym_09c_nifti.nii",
            }]
        }


    def test_space_init(self):
        space = Space(self.space_id, self.name, self.url, self.ziptarget)
        self.assertIsNotNone(space)

    def test_space_from_json_with_zip(self):
        space = Space.from_json(self.json_space_with_zip)
        self.assertEqual(
            str(space),
            self.name
        )
        self.assertEqual(len(space.volume_src),1)
        vsrc = space.volume_src[0]     
        self.assertEqual(space.type,self.ttype)
        self.assertEqual(vsrc.zipped_file,self.ziptarget)

    def test_space_from_json_without_zip(self):
        space = Space.from_json(self.json_space_without_zip)
        self.assertEqual(
            str(space),
            self.name
        )
        self.assertEqual(len(space.volume_src),1)
        vsrc = space.volume_src[0]     
        self.assertIsNone(vsrc.zipped_file)

    def test_space_registry(self):
        spaces = sb.atlases['human'].spaces
        self.assertEqual(len(spaces), 4)


if __name__ == "__main__":
    unittest.main()



# Copyright 2018-2020 Institute of Neuroscience and Medicine (INM-1),
# Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import logger
from . import retrieval 
import os
import requests
import json
import numpy as np
from cloudvolume import CloudVolume,Bbox
from memoization import cached
import nibabel as nib

@cached
def is_ngprecomputed(url):
    # Check if the given URL is likely a neuroglancer precomputed cloud store
    try: 
        r = requests.get(url+"/info")
        info = json.loads(r.content)
        return info['type'] in ['image','segmentation']
    except Exception as _:
        return False


@cached
def load_ngprecomputed(url,resolution):
    """
    Shortcut for loading data from a neuroglancer precomputed volume directly
    into a spatial image object
    """
    V = BigBrainVolume(url)
    return V.build_image(resolution,clip=True)

def bbox3d(A):
    # https://stackoverflow.com/questions/31400769/bounding-box-of-numpy-array
    r = np.any(A, axis=(1, 2))
    c = np.any(A, axis=(0, 2))
    z = np.any(A, axis=(0, 1))
    rmin, rmax = np.where(r)[0][[0, -1]]
    cmin, cmax = np.where(c)[0][[0, -1]]
    zmin, zmax = np.where(z)[0][[0, -1]]
    return np.array([
        [rmin, rmax], 
        [cmin, cmax], 
        [zmin, zmax],
        [1,1]
    ])


class BigBrainVolume:
    """
    TODO use siibra requests cache
    
    """
    # function to switch x/y coordinates on a vector or matrix.
    # Note that direction doesn't matter here since the inverse is the same.
    switch_xy = lambda X : np.dot(np.identity(4)[[1,0,2,3],:],X) 

    # Gigabyte size that is considered feasible for ad-hoc downloads of
    # BigBrain data. This is used to avoid accidental huge downloads.
    gbyte_feasible = 0.5
    
    def __init__(self,ngsite,fill_missing=True):
        """
        ngsite: base url of neuroglancer http location
        """
        with requests.get(ngsite+'/transform.json') as r:
            self._translation_nm = np.array(json.loads(r.content))[:,-1]
        with requests.get(ngsite+'/info') as r:
            self.info = json.loads(r.content)
        self.volume = CloudVolume(ngsite,fill_missing=fill_missing,progress=False)
        self.ngsite = ngsite
        self.nbits = np.iinfo(self.volume.info['data_type']).bits
        self.bbox_phys = self._bbox_phys()
        self.resolutions_available = {
                np.min(v['resolution'])/1000 : {
                    'mip':i,
                    'GBytes':np.prod(v['size'])*self.nbits/(8*1024**3)
                    }
                for i,v in enumerate(self.volume.scales) }
        self.helptext = "\n".join(["{:7.0f} micron {:10.4f} GByte".format(k,v['GBytes']) 
            for k,v in self.resolutions_available.items()])

    def largest_feasible_resolution(self):
        # returns the highest resolution in micrometer that is available and
        # still below the threshold of downloadable volume sizes.
        return min([res
            for res,v in self.resolutions_available.items()
            if v['GBytes']<self.gbyte_feasible ])
        
    def affine(self,mip,clip=False):
        """
        Builds the affine matrix that maps voxels 
        at the given mip to physical space in mm.
        Parameters:
        -----------
        clip : Boolean, or Bbox
            If true, clip by computing the bounding box from nonempty pixels
            if False, get the complete data of the selected mip
            If Bbox, clip by this bounding box
        """

        # correct clipping offset, if needed
        voxelshift = np.identity(4)
        if (type(clip)==bool) and clip is True:
            voxelshift[:3,-1] = self._clipcoords(mip)[:3,0]
        elif isinstance(clip,Bbox):
            voxelshift[:3,-1] = clip.minpt

        # retrieve the pixel resolution
        resolution_nm = self.info['scales'][mip]['resolution']

        # build affine matrix in nm physical space
        affine = np.identity(4)
        for i in range(3):
            affine[i,i] = resolution_nm[i]
            affine[i,-1] = self._translation_nm[i]
            
        # warp from nm to mm   
        affine[:3,:]/=1000000.
    
        return np.dot(affine,voxelshift)
        #return BigBrainVolume.switch_xy(np.dot(affine,voxelshift))

    def _clipcoords(self,mip):
        # compute clip coordinates in voxels for the given mip 
        # from the pre-computed physical bounding box coordinates
        
        logger.debug("Computing bounding box coordinates at mip {}".format(mip))
        phys2vox = np.linalg.inv(self.affine(mip))
        clipcoords = np.dot(phys2vox,self.bbox_phys).astype('int')
        # clip bounding box coordinates to actual shape of the mip
        clipcoords[:,0] = np.maximum(clipcoords[:,0],0)
        clipcoords[:,1] = np.minimum(clipcoords[:,1],self.volume.mip_shape(mip))
        return clipcoords

    def _load_data(self,mip,clip=False,force=False):
        """
        Actually load image data.
        TODO: Check amount of data beforehand and raise an Exception if it is over a reasonable threshold.
        NOTE: this function caches chunks as numpy arrays (*.npy) to the
        CACHEDIR defined in the retrieval module.
        
        Parameters:
        -----------
        clip : Boolean, or Bbox
            If true, clip by computing the bounding box from nonempty pixels
            if False, get the complete data of the selected mip
            If Bbox, clip by this bounding box
        force : Boolean (default: False)
            if true, will start downloads even if they exceed the download
            threshold set in the gbytes_feasible member variable.
        """
        if (type(clip)==bool) and clip is True:
            clipcoords = self._clipcoords(mip)
            bbox = Bbox(clipcoords[:3,0],clipcoords[:3,1])
        elif isinstance(clip,Bbox):
            # make sure the bounding box is integer, some are not
            bbox = Bbox(
                    np.array(clip.minpt).astype('int'),
                    np.array(clip.maxpt).astype('int'))
        else:
            bbox = Bbox([0, 0, 0],self.volume.mip_shape(mip))
        gbytes = bbox.volume()*self.nbits/(8*1024**3)
        if not force and gbytes>BigBrainVolume.gbyte_feasible:
            # TODO would better do an estimate of the acutal data size
            logger.error("Data request is too large (would result in an ~{:.2f} GByte download, the limit is {}).".format(gbytes,self.gbyte_feasible))
            print(self.helptext)
            raise RuntimeError("The requested resolution is too high to provide a feasible download, but you can override this behavior with the 'force' parameter.")
        cachefile = retrieval.cachefile("{}{}{}".format(
            self.ngsite, bbox.serialize(), str(mip)).encode('utf8'),suffix='npy')
        if os.path.exists(cachefile):
            return np.load(cachefile)
        else:
            data = self.volume.download(bbox=bbox,mip=mip)
            np.save(cachefile,np.array(data))
            return np.array(data)

    def determine_mip(self,resolution=None):
        # given a resolution in micrometer, try to determine the mip that can
        # be used to move on.
        if resolution is None:
            maxres = self.largest_feasible_resolution()
            logger.info('Using the largest feasible resolution of {} micron'.format(maxres))
            return self.resolutions_available[maxres]['mip']
        elif resolution in self.resolutions_available.keys(): 
            return self.resolutions_available[resolution]['mip']
        logger.error('The requested resolution ({} micron) is not available. Choose one of:'.format(resolution))
        print(self.helptext)
        return None
        
    def build_image(self,resolution,clip=True,transform=lambda I: I, force=False):
        """
        Compute and return a spatial image for the given mip.
        
        Parameters:
        -----------
        clip : Boolean, or Bbox
            If true, clip by computing the bounding box from nonempty pixels
            if False, get the complete data of the selected mip
            If Bbox, clip by this bounding box
        force : Boolean (default: False)
            If true, will start downloads even if they exceed the download
            threshold set in the gbytes_feasible member variable.
        """
        mip = self.determine_mip(resolution)
        if not mip:
            raise ValueError("Invalid image resolution for this neuroglancer precomputed tile source.")
        return nib.Nifti1Image(
            transform(self._load_data(mip,clip,force)),
            affine=self.affine(mip,clip)
        )
    
    def _enclosing_chunkgrid(self, mip, bbox_phys):
        """
        Produce grid points representing the chunks of the mip 
        which enclose a given bounding box. The bounding box is given in
        physical coordinates, but the grid is returned in voxel spaces of the
        given mip.
        """
        
        # some helperfunctions to produce the smallest range on a grid enclosing another range 
        cfloor = lambda x,s: int(np.floor(x/s)*s)
        cceil = lambda x,s: int(np.ceil(x/s)*s)+1
        crange = lambda x0,x1,s: np.arange(cfloor(x0,s),cceil(x1,s),s)
        
        # project the bounding box to the voxel grid of the selected mip
        bb = np.dot(np.linalg.inv(self.affine(mip)),bbox_phys)
        
        # compute the enclosing chunk grid
        chunksizes = self.volume.scales[mip]['chunk_sizes'][0]
        x,y,z = [crange(bb[i][0],bb[i][1],chunksizes[i]) 
                 for i in range(3)]
        xx,yy,zz = np.meshgrid(x,y,z)
        return np.vstack([xx.ravel(),yy.ravel(),zz.ravel(),zz.ravel()*0+1])

    def _bbox_phys(self):
        """
        Estimates the bounding box of the nonzero values 
        in the data volume, in physical coordinates. 
        Estimation is done from the lowest resolution for 
        efficiency, so it is not fully accurate.
        """
        volume = self._load_data(-1,clip=False)
        affine = self.affine(-1,clip=False)
        bbox_vox = bbox3d(volume)
        return np.dot(affine,bbox_vox)






# Copyright 2018-2020 Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import logger,arrays, QUIET
from .space import Space,SpaceVOI
from .commons import MapType,ParcellationIndex
from .volumesrc import ImageProvider,VolumeSrc
from .arrays import create_homogeneous_array,create_gaussian_kernel,argmax_dim4
from .region import Region

import numpy as np
import nibabel as nib
from nilearn import image
from memoization import cached
from tqdm import tqdm
from abc import abstractmethod
from typing import Union

# Which types of available volumes should be preferred if multiple choices are available?
PREFERRED_VOLUMETYPES = ['nii','neuroglancer/precomputed','detailed maps']

def create_map(parcellation, space:Space, maptype:MapType ):
    """
    Creates a new ParcellationMap object of the given type.
    """
    classes = {
        MapType.LABELLED:LabelledParcellationMap,
        MapType.CONTINUOUS:ContinuousParcellationMap
    }
    if maptype in classes:
        obj = classes[maptype](parcellation,space)
    elif maptype is None:
        logger.warning('No maptype provided when requesting the parcellation map. Falling back to MapType.LABELLED')
        obj = classes[MapType.LABELLED](parcellation,space)
    else:
        raise ValueError(f"Invalid maptype: '{maptype}'")
    if len(obj)==0:
        raise ValueError(f"No data found to construct a {maptype} map for {parcellation.name} in {space.name}.")
    
    return obj

class ParcellationMap(ImageProvider):
    """
    Represents a brain map in a particular reference space, with
    explicit knowledge about the region information per labelindex or channel.

    There are two types:
        1) Parcellation maps / labelled volumes (MapType.LABELLED)
            A 3D or 4D volume with integer labels separating different,
            non-overlapping regions. The number of regions corresponds to the
            number of nonzero image labels in the volume.
        2) 4D overlapping regional maps (often probability maps) (MapType.CONTINUOUS)
            a 4D volume where each "time"-slice is a 3D volume representing
            a map of a particular brain region. This format is used for
            probability maps and similar continuous forms. The number of
            regions correspond to the z dimension of the 4 object.

    ParcellationMaps can be also constructred from neuroglancer (BigBrain) volumes if
    a feasible downsampled resolution is provided.
    """
    _regions_cached = None
    _maploaders_cached = None

    def __init__(self, parcellation, space: Space, maptype=MapType ):
        """
        Construct a ParcellationMap for the given parcellation and space.

        Parameters
        ----------
        parcellation : Parcellation
            The parcellation object used to build the map
        space : Space
            The desired template space to build the map
        maptype : MapType
            The desired type of the map
        """
        if not parcellation.supports_space(space):
            raise ValueError( 'Parcellation "{}" does not provide a map for space "{}"'.format(
                parcellation.name, space.name ))

        self.maptype = maptype
        self.parcellation = parcellation
        self.space = space

    @property
    def maploaders(self):
        if self._maploaders_cached is None:
            self._define_maps_and_regions()
        return self._maploaders_cached

    @property
    def regions(self):
        """
        Dictionary of regions associated to the parcellion map, indexed by ParcellationIndex.
        Lazy implementation - self._link_regions() will be called when the regions are accessed for the first time.
        """
        if self._regions_cached is None:
            self._define_maps_and_regions()
        return self._regions_cached

    @property
    def names(self):
        return self.parcellation.names

    @abstractmethod
    def _define_maps_and_regions(self):
        """
        implemented by derived classes, to produce the lists _regions_cached and _maploaders_cached. 
        The first is a dictionary indexed by ParcellationIndex, 
        the latter a list of functions for loading the different maps. 
        """
        pass

    def fetchall(self,resolution_mm=None,voi:SpaceVOI=None):
        """
        Returns an iterator to fetch all available maps sequentially:

        Parameters
        ----------
        resolution_mm : float or None (optional)
            Physical resolution of the map, used for multi-resolution image volumes. 
            If None, the smallest possible resolution will be chosen. 
            If -1, the largest feasible resolution will be chosen.        
        """
        logger.debug(f'Iterator for fetching {len(self)} parcellation maps')
        return (fnc(res=resolution_mm,voi=voi) for fnc in self.maploaders)

    def fetch(self,resolution_mm:float=None,mapindex:int=0, voi:SpaceVOI=None):
        """
        Fetches the actual image data

        Parameters
        ----------
        resolution_mm : float or None (optional)
            Physical resolution of the map, used for multi-resolution image volumes. 
            If None, the smallest possible resolution will be chosen. 
            If -1, the largest feasible resolution will be chosen.  
        mapindex : int
            The index of the available maps to be fetched.       
        """
        if mapindex<len(self):
            return self.maploaders[mapindex](res=resolution_mm,voi=voi)
        else:
            raise ValueError(f"'{len(self)}' maps available, but a mapindex of {mapindex} was requested.")

    def _load_regional_map(self, region:Region, resolution_mm, voi:SpaceVOI=None, clip:bool=False):
        logger.debug(f"Loading regional map for {region.name} in {self.space.name}")
        rmap = region.get_regional_map(self.space, self.maptype).fetch(resolution_mm=resolution_mm,voi=voi,clip=clip)
        return rmap

    def __len__(self):
        """
        Returns the number of maps available in this parcellation.
        """
        return len(self.maploaders)

    def __contains__(self,spec):
        """
        Test if a 3D map identified by the given specification is included in this parcellation map. 
        For integer values, it is checked wether a corresponding slice along the fourth dimension could be extracted.
        Alternatively, a region object can be provided, and it will be checked wether the region is mapped.
        You might find the decode_region() function of Parcellation and Region objects useful for the latter.
        """
        if isinstance(spec,int):
            return spec in range(len(self.maploaders))
        elif isinstance(spec,Region):
            for _,region in self.regions.items():
                if region==spec:
                    return True
        return False

    def decode_label(self,mapindex=None,labelindex=None):
        """
        Decode the region associated to a particular index.

        Parameters
        ----------
        mapindex : Sequential index of the 3D map used, if more than one are included
        labelindex : Label index of the region, if the map is a labelled volume 
        """
        pindex = ParcellationIndex(map=mapindex,label=labelindex)
        region = self.regions.get(pindex)
        if region is None:
            raise ValueError(f"Could not decode parcellation index {pindex}")
        else:
            return region

    
    def decode_region(self,regionspec:Union[str,Region]):
        """
        Find the ParcellationIndex for a given region.

        Parameters
        ----------
        regionspec : str or Region
            Partial name of region, or Region object

        Return
        ------
        list of MapIndex objects
        """
        region = self.parcellation.decode_region(regionspec) if isinstance(regionspec,str) else regionspec
        result = []
        for idx,r in self.regions.items():
            if r==region:
                return [idx] 
            elif r.has_parent(region):
                result.append(idx)
        if len(result)==0:
            raise IndexError(f"Could not decode region specified by {regionspec} in {self.parcellation.name}")
        return result

    def extract_regionmap(self,regionspec:Union[str,int,Region],resolution_mm=None,voi:SpaceVOI=None):
        """
        Extract the mask for one particular region. For parcellation maps, this
        is a binary mask volume. For overlapping maps, this is the
        corresponding slice, which typically is a volume of float type.

        Parameters
        ----------
        regionspec : labelindex, partial region name, or Region
            The desired region.
        resolution_mm : float or None (optional)
            Physical resolution of the map, used for multi-resolution image volumes. 
            If None, the smallest possible resolution will be chosen. 
            If -1, the largest feasible resolution will be chosen. 

        Return
        ------
        Nifti1Image, if found, otherwise None
        """
        indices = self.decode_region(regionspec)
        mapimgs = []
        for index in indices:
            mapimg = self.fetch(resolution_mm=resolution_mm, mapindex=index.map,voi=voi)
            if index.label is not None:
                mapimg = nib.Nifti1Image(
                    dataobj=(mapimg.get_fdata()==index.label).astype(np.uint8),
                    affine=mapimg.affine)
            mapimgs.append(mapimg)

        if len(mapimgs)==1:
            return mapimgs[0]
        elif self.maptype==MapType.LABELLED:
            m = mapimgs[0]
            for m2 in mapimgs[1:]:
                m.dataobj[m2.dataobj>0]=1
            return m
        else:
            logger.info(f"4D volume with {len(mapimgs)} continuous region maps extracted from region specification '{regionspec}'")
            return image.concat_imgs(mapimgs)
    
    def get_shape(self,resolution_mm=None):
        return list(self.space.get_template().get_shape()) + [len(self)]

    def is_float(self):
        return self.maptype==MapType.CONTINUOUS

class LabelledParcellationMap(ParcellationMap):
    """
    Represents a brain map in a reference space, with
    explicit knowledge about the region information per labelindex or channel.
    Contains a Nifti1Image object as the "image" member.

    This form defines parcellation maps / labelled volumes (MapType.LABELLED),
    A 3D or 4D volume with integer labels separating different,
    non-overlapping regions. The number of regions corresponds to the
    number of nonzero image labels in the volume.
    """

    def __init__(self, parcellation, space: Space ):
        """
        Construct a ParcellationMap for the given parcellation and space.

        Parameters
        ----------
        parcellation : Parcellation
            The parcellation object used to build the map
        space : Space
            The desired template space to build the map
        """
        super().__init__(parcellation, space,MapType.LABELLED)

    def _define_maps_and_regions(self):

        self._maploaders_cached=[]
        self._regions_cached = {}

        # determine the map loader functions for each available map
        for mapname in self.parcellation.volume_src[self.space]:

            # Determine the preferred volume source for loading the parcellation map
            volume_sources = sorted(
                    self.parcellation.volume_src[self.space][mapname],
                    key=lambda vsrc: PREFERRED_VOLUMETYPES.index(vsrc.volume_type))
            if len(volume_sources)==0:
                logger.error(f'No suitable volume source for {self.parcellation.name} in {self.space.name}')
                continue
            source = volume_sources[0]

            # Choose map loader function and populate label-to-region maps
            if source.volume_type=="detailed maps":
                self._maploaders_cached.append(lambda res=None,voi=None: self._collect_maps(resolution_mm=res,voi=voi))
                # collect all available region maps to maps label indices to regions
                regions = [r for r in self.parcellation.regiontree 
                        if r.has_regional_map(self.space,MapType.LABELLED)]
                for region in regions:
                    assert(region.index.label)
                    self._regions_cached[ParcellationIndex(map=0,label=region.index.label)] = region
            elif source.volume_type==self.space.type:
                self._maploaders_cached.append(lambda res=None,s=source,voi=None: self._load_map(s,resolution_mm=res,voi=voi))
                # load map at lowest resolution to map label indices to regions
                mapindex = len(self._maploaders_cached)-1
                with QUIET:
                    m = self._maploaders_cached[mapindex](res=None)
                unmatched = []
                for labelindex in np.unique(m.get_fdata()):
                    if labelindex!=0:
                        pindex = ParcellationIndex(map=mapindex,label=labelindex)
                        try:
                            region = self.parcellation.decode_region(pindex)
                            if labelindex>0:
                                self._regions_cached[pindex] = region
                        except ValueError:
                            unmatched.append(pindex)
                if unmatched:
                    logger.warning(f"{len(unmatched)} parcellation indices in labelled volume couldn't be matched to region definitions in {self.parcellation.name}")
            
    @cached
    def _load_map(self,volume_src:VolumeSrc,resolution_mm:float,voi:SpaceVOI):
        m = volume_src.fetch(resolution_mm=resolution_mm,voi=voi)
        if len(m.dataobj.shape)==4 and m.dataobj.shape[3]>1:
            logger.info(f"{m.dataobj.shape[3]} continuous maps given - using argmax to generate a labelled volume. ")
            m = argmax_dim4(m)
        if m.dataobj.dtype.kind=='f':
            logger.warning(f"Floating point image type encountered when building a labelled volume for {self.parcellation.name}, converting to integer.")
            m = nib.Nifti1Image(dataobj=np.asarray(m.dataobj,dtype=int),affine=m.affine)
        return m

    @cached
    def _collect_maps(self,resolution_mm,voi):
        """
        Build a 3D volume from the list of available regional maps.

        Return
        ------
        Nifti1Image, or None if no maps are found.
        
        """
        m = None

        # generate empty mask covering the template space
        tpl = self.space.get_template().fetch(resolution_mm,voi=voi)
        m = nib.Nifti1Image(np.zeros_like(tpl.dataobj,dtype='uint'),tpl.affine)

        # collect all available region maps
        regions = [r for r in self.parcellation.regiontree 
                if r.has_regional_map(self.space,MapType.LABELLED)]

        msg =f"Loading {len(regions)} regional maps for space '{self.space.name}'..."
        logger.info(msg)
        for region in regions:
            assert(region.index.label)
            # load region mask
            mask_ = self._load_regional_map(region,resolution_mm=resolution_mm,voi=voi)
            if not mask_:
                continue
            # build up the aggregated mask with labelled indices
            if mask_.shape!=m.shape:
                mask = image.resample_to_img(mask_,m,interpolation='nearest')
            else:
                mask = mask_
            m.dataobj[mask.dataobj>0] = region.index.label
            self._regions_cached[ParcellationIndex(map=0,label=region.index.label)] = region

        return m


class ContinuousParcellationMap(ParcellationMap):
    """
    Represents a brain map in a particular reference space, with
    explicit knowledge about the region information per labelindex or channel.

    This form represents overlapping regional maps (often probability maps) (MapType.CONTINUOUS)
    where each "time"-slice is a 3D volume representing
    a map of a particular brain region. This format is used for
    probability maps and similar continuous forms. The number of
    regions correspond to the z dimension of the 4 object.
    """

    def __init__(self, parcellation, space: Space ):
        """
        Construct a ParcellationMap for the given parcellation and space.

        Parameters
        ----------
        parcellation : Parcellation
            The parcellation object used to build the map
        space : Space
            The desired template space to build the map
        """
        super().__init__(parcellation, space, MapType.CONTINUOUS)

    def _define_maps_and_regions(self):
        self._maploaders_cached=[]
        self._regions_cached={}

        # check for maps associated to the parcellations
        for mapname in self.parcellation.volume_src[self.space]:

            # Multiple volume sources could be given - find the preferred one
            volume_sources = sorted(
                    self.parcellation.volume_src[self.space][mapname],
                    key=lambda vsrc: PREFERRED_VOLUMETYPES.index(vsrc.volume_type))
            if len(volume_sources)==0:
                logger.error(f'No suitable volume source for "{mapname}"' +
                             f'of {self.parcellation.name} in {self.space.name}')
                continue
            source = volume_sources[0]
            
            if not all([source.is_float(),source.is_4D()]):
                continue
            if source.get_shape()[3]<2:
                continue

            # The source is 4D float, that's what we are looking for. 
            # We assume the fourth dimension contains the regional continuous maps.
            nmaps = source.get_shape()[3]
            logger.info(f'{nmaps} continuous maps will be extracted from 4D volume for {self.parcellation}.')
            for i in range(nmaps):
                self._maploaders_cached.append(
                    lambda res=None,voi=None,mi=i: source.fetch(resolution_mm=res,voi=voi,mapindex=mi))
                region = self.parcellation.decode_region(i+1)
                pindex = ParcellationIndex(map=i,label=None)
                self._regions_cached[pindex] = region

            # we are finished, no need to look for regional map.
            return
                
        # otherwise we look for continuous maps associated to individual regions
        regions = [r for r in self.parcellation.regiontree 
                if r.has_regional_map(self.space,MapType.CONTINUOUS)]
        logger.info(f'{len(regions)} regional continuous maps found for {self.parcellation}.')
        for i,region in enumerate(regions):
            self._maploaders_cached.append(
                lambda r=region,res=None,voi=None:self._load_regional_map(r,resolution_mm=res,voi=voi))
            if region in self.regions.values():
                logger.debug(f"Region already seen in tree: {region.key}")
            pindex = ParcellationIndex(map=i,label=None)
            self._regions_cached[pindex] = region

    @cached
    def assign_coordinates(self,xyz_phys,sigma_mm=1,sigma_truncation=3):
        """
        Assign regions to a physical coordinates with optional standard deviation.

        Parameters
        ----------
        xyz_phys : 3D point(s) in physical coordinates of the template space of the ParcellationMap
            Can be one 3D coordinate tuple, list of 3D tuples, Nx3 or Nx4 array of coordinate tuples, 
            str of the form "3.1mm, -3.1mm, 80978mm", or list of such strings.
            See arrays.create_homogeneous_array
        sigma_mm : float (default: 1)
            standard deviation /expected localization accuracy of the point, in
            mm units. A 3D Gaussian distribution with that
            bandwidth will be used for representing the location.
        sigma_truncation : float (default: 3)
            If sigma_phys is nonzero, this factor is used to determine where to
            truncate the Gaussian kernel in standard error units.
        """
        assert(sigma_mm>=1)

        # Convert input to Nx4 list of homogenous coordinates
        assert(len(xyz_phys)>0)
        XYZH = create_homogeneous_array(xyz_phys)
        numpts = XYZH.shape[0]

        # convert sigma to voxel coordinates
        tpl = self.space.get_template().fetch()
        phys2vox = np.linalg.inv(tpl.affine)
        scaling = np.array([np.linalg.norm(tpl.affine[:,i]) for i in range(3)]).mean()
        sigma_vox = sigma_mm/scaling

        assignments = []
        if sigma_vox<3:
            N = len(self)
            msg = f"Assigning {numpts} coordinates to {N} maps"
            assignments = [[] for n in range(numpts)]
            for mapindex,loadfnc in tqdm(enumerate(self.maploaders),total=len(self),desc=msg,unit=" maps"):
                pmap = loadfnc()
                p2v = np.linalg.inv(tpl.affine)
                A = pmap.get_fdata()
                pindex = ParcellationIndex(map=mapindex,label=None)
                region = self.decode_label(mapindex=mapindex)
                for i,xyzh in enumerate(XYZH):
                    x,y,z = (np.dot(p2v,xyzh)+.5).astype('int')[:3]
                    value = A[x,y,z]
                    if value>0:
                        assignments[i].append((region,pmap,value))
        else:
            logger.info((f"Assigning {numpts} uncertain coordinates (stderr={sigma_mm}) to {len(self)} maps." ))
            kernel = create_gaussian_kernel(sigma_vox,sigma_truncation)
            r = int(kernel.shape[0]/2) # effective radius
            for i,xyzh in enumerate(XYZH):
                xyz_vox = (np.dot(phys2vox,xyzh)+.5).astype('int')
                shift = np.identity(4)
                shift[:3,-1] = xyz_vox[:3]-r
                W = nib.Nifti1Image(dataobj=kernel,affine=np.dot(tpl.affine,shift))
                assignments.append(self.assign(W,msg=", ".join([f"{v:.1f}" for v in xyzh[:3]])))

        return assignments
    
    def assign(self,img:nib.Nifti1Image,msg=None,quiet=False):
        """
        Assign the region of interest represented by a given volumetric image to continuous brain regions in this map.

        Parameters:
        -----------
        img : Nifti1Image
            The input region of interest, typically a binary mask or statistical map.
        msg : str, default:None
            Message to display with the progress bar
        quiet: Boolen, default:False
            If true, no progess indicator will be displayed
        """

        if msg is None and not quiet:
            msg=f"Assigning structure to {len(self)} maps"

        # How to visualize progress from the iterator?
        progress = lambda f: f
        if not quiet: 
            progress = lambda f: tqdm(f,total=len(self),desc=msg,unit="maps")

        # setup assignment loop
        values = {}
        pmaps = {}

        for mapindex,loadfnc in progress(enumerate(self.maploaders)):

            # load the regional map
            this = loadfnc()
            if not this:
                logger.warning(f"Could not load regional map for {self.regions[mapindex].name}")
                continue

            scores = compare_maps(img,this)
            if scores['overlap']>0:
                pmaps[mapindex] = this
                values[mapindex] = scores

        assignments = [(self.decode_label(mapindex=i),pmaps[i],value) 
                    for i,value in sorted(
                        values.items(),
                        key=lambda item:abs(item[1]['correlation']),reverse=True)]
        return assignments


# getting nonzero pixels of pmaps is one of the most time consuming tasks when computing metrics, 
# so we cache the nonzero coordinates of array objects at runtime.
NZCACHE = {}
def nonzero_coordinates(arr):
    obj_id = id(arr)
    if obj_id not in NZCACHE:
        NZCACHE[obj_id] = np.c_[np.nonzero(arr>0)]
    return NZCACHE[obj_id]

def compare_maps(map1:nib.Nifti1Image,map2:nib.Nifti1Image):
    """
    Compare two maps, given as Nifti1Image objects.
    This function exploits that nibabel's get_fdata() caches the numerical arrays, 
    so we can use the object id to cache extraction of the nonzero coordinates.
    Repeated calls involving the same map will therefore be much faster as they 
    will only access the image array if overlapping pixels are detected.

    It is recommended to install the indexed-gzip package, 
    which will further speed this up.
    """
    
    a1,a2 = [m.get_fdata().squeeze() for m in (map1,map2)]
    
    homog = lambda XYZ : np.c_[XYZ,np.ones(XYZ.shape[0])]
    colsplit = lambda XYZ: np.split(XYZ,3,axis=1)

    # Compute the nonzero voxels in map2 and their correspondences in map1
    XYZnz2 = nonzero_coordinates(a2)
    N2 = XYZnz2.shape[0]
    warp2on1 = np.dot(np.linalg.inv(map1.affine),map2.affine)
    XYZnz2on1 = (np.dot(warp2on1,homog(XYZnz2).T).T[:,:3]+.5).astype('int')

    # valid voxel pairs
    valid = np.all(
        np.logical_and.reduce([
            XYZnz2on1>=0, XYZnz2on1<map1.shape[:3],
            XYZnz2>=0, XYZnz2<map2.shape[:3]]), 1)
    X1,Y1,Z1 = colsplit(XYZnz2on1[valid,:])
    X2,Y2,Z2 = colsplit(XYZnz2[valid,:])

    # intersection
    v1,v2 = a1[X1,Y1,Z1].squeeze(),a2[X2,Y2,Z2].squeeze()
    m1,m2 = ((_>0).astype('uint8') for _ in [v1,v2])
    intersection = np.minimum(m1,m2).sum()
    if intersection==0:
            return {'overlap':0,'contained':0,'contains':0,'correlation':0}
    
    # Compute the nonzero voxels in map1 with their correspondences in map2
    XYZnz1 = nonzero_coordinates(a1)
    N1 = XYZnz1.shape[0]
    warp1on2 = np.dot(np.linalg.inv(map2.affine),map1.affine)
    
    # Voxels referring to the union of the nonzero pixels in both maps
    XYZa1 = np.unique(np.concatenate((XYZnz1,XYZnz2on1)),axis=0)    
    XYZa2 = (np.dot(warp1on2,homog(XYZa1).T).T[:,:3]+.5).astype('int')
    valid = np.all(
        np.logical_and.reduce([
            XYZa1>=0, XYZa1<map1.shape[:3],
            XYZa2>=0, XYZa2<map2.shape[:3]]), 1)
    Xa1,Ya1,Za1 = colsplit(XYZa1[valid,:])
    Xa2,Ya2,Za2 = colsplit(XYZa2[valid,:])
    
    # pearson's r wrt to full size image
    x = a1[Xa1,Ya1,Za1].squeeze()
    y = a2[Xa2,Ya2,Za2].squeeze()
    mu_x = x.sum()/a1.size
    mu_y = y.sum()/a2.size
    x0 = x-mu_x
    y0 = y-mu_y
    r = np.sum(np.multiply(x0,y0)) / np.sqrt(np.sum(x0**2)*np.sum(y0**2))
    
    bx = (x>0).astype('uint8')
    by = (y>0).astype('uint8')

    return {
        'overlap':      intersection/np.maximum(bx,by).sum(),
        'contained':    intersection/N1,
        'contains':     intersection/N2,
        'correlation':  r
    }


import os
import unittest
from siibra.atlas import REGISTRY
import siibra as sb
from test.get_token import get_token

token = get_token()
os.environ['HBP_AUTH_TOKEN'] = token["access_token"]

class TestSelectionBrainRegions(unittest.TestCase):

    def test_select_brain_regions(self):
        atlas = REGISTRY.MULTILEVEL_HUMAN_ATLAS
        atlas.select_parcellation('2.9')
        # we can just give a string and see if the system can disambiguiate it
        atlas.select_region('v1')
        print("Selected region from 'v1' is", atlas.selected_region)
        self.assertEqual(atlas.selected_region.name, 'Area hOc1 (V1, 17, CalcS)')
        self.assertTrue(len(atlas.selected_region.children) == 2)

        print('v1 includes the left and right hemisphere!')
        print(repr(atlas.selected_region))

        # we can be more specific easily
        atlas.select_region('v1 left')
        print("Selected region from 'v1 left' is", atlas.selected_region)
        self.assertEqual(atlas.selected_region.name, 'Area hOc1 (V1, 17, CalcS) left')
        self.assertTrue(len(atlas.selected_region.children) == 0)

        # we can also auto-complete on the 'regionnames' attribute of the atlas
        # - this immediately leads to a unique selection
        atlas.select_region(atlas.regionnames.AREA_HOC1_V1_17_CALCS_LEFT_HEMISPHERE)
        self.assertEqual(atlas.selected_region.name, 'Area hOc1 (V1, 17, CalcS) left')
        self.assertTrue(len(atlas.selected_region.children) == 0)


if __name__ == "__main__":
    unittest.main()



import os
import unittest
from siibra.atlas import REGISTRY
from siibra.features import modalities
import siibra as sb
from siibra import retrieval
from test.get_token import get_token

token = get_token()
os.environ['HBP_AUTH_TOKEN'] = token["access_token"]

class TestExtractTransmitterReceptorDensities(unittest.TestCase):

    def test_extract_densities(self):
        atlas = sb.atlases['human']
        atlas.select_parcellation('2.9')
        atlas.select_region('hoc1 left')
        features = atlas.get_features(modalities.ReceptorDistribution)
        self.assertTrue(len(features) == 1)
        self.assertEqual(features[0].name, 'Density measurements of different receptors for Area hOc1 (V1, 17, CalcS) [human, v1.0]')

        regions = ['hOc1', 'hOc2', 'IFG']
        query = sb.features.receptors.ReceptorQuery(atlas)
        self.assertTrue(len(query.features) >= 42)
        for q in regions:
            matched_features = [f for f in query.features if f.region.matches(q)]
            self.assertGreater(len(matched_features),0)


if __name__ == "__main__":
    unittest.main()



# Copyright 2018-2020 Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import logger, spaces, volumesrc, parcellationmap
from .space import Space
from .region import Region
from .config import ConfigurationRegistry
from .commons import create_key,MapType,ParcellationIndex,HasOriginDataInfo,OriginDataInfo
from typing import Union
import numpy as np
import nibabel as nib
from nilearn import image
from tqdm import tqdm
from memoization import cached

class ParcellationVersion:
    def __init__(self, name=None, collection=None, prev_id=None, next_id=None, deprecated=False):
        self.name=name
        self.collection=collection
        self.next_id=next_id
        self.prev_id=prev_id
        self.deprecated=deprecated

    def __eq__(self,other):
        return all([
            self.name==other.name,
            self.collection==other.collection])
    
    def __str__(self):
        return self.name

    def __repr__(self):
        return  self.name

    def __iter__(self):
        yield 'name', self.name
        yield 'prev', self.prev.id if self.prev is not None else None
        yield 'next', self.next.id if self.next is not None else None
        yield 'deprecated', self.deprecated

    def __lt__(self,other):
        """ < operator, useful for sorting by version"""
        successor = self.next
        while successor is not None:
            if successor.version==other:
                return True
            successor = successor.version.next
        return False

    @property
    def next(self):
        if self.next_id is None:
            return None
        try:
            return REGISTRY[self.next_id]
        except IndexError:
            return None
        except NameError:
            logger.warning('Accessing REGISTRY before its declaration!')
    
    @property
    def prev(self):
        if self.prev_id is None:
            return None
        try:
            return REGISTRY[self.prev_id]
        except IndexError:
            return None
        except NameError:
            logger.warning('Accessing REGISTRY before its declaration!')

    @staticmethod
    def from_json(obj):
        """
        Provides an object hook for the json library to construct a
        ParcellationVersion object from a json string.
        """
        if obj is None:
            return None
        return ParcellationVersion(
            obj.get('name', None), 
            obj.get('collectionName',None),
            prev_id=obj.get('@prev', None), 
            next_id=obj.get('@next', None),
            deprecated=obj.get('deprecated', False))

class Parcellation(HasOriginDataInfo):

    _regiontree_cached = None

    def __init__(self, identifier : str, name : str, version=None, modality=None, regiondefs=[]):
        """
        Constructs a new parcellation object.

        Parameters
        ----------
        identifier : str
            Unique identifier of the parcellation
        name : str
            Human-readable name of the parcellation
        version : str or None
            a version specification, optional
        modality  :  str or None
            a specification of the modality used for creating the parcellation.
        """
        HasOriginDataInfo.__init__(self)
        self.id = identifier
        self.name = name
        self.key = create_key(name)
        self.version = version

        # TODO replace by instanced methods in HasOriginDataInfo
        self.publications = []
        self.description = ""

        self.volume_src = {}
        self.modality = modality
        self._regiondefs = regiondefs

        # If set, thresholded continuous maps will be preferred
        # over static labelled maps for building and using region masks.
        # This will influence the shape of region masks used for filtering.
        self.continuous_map_threshold = None

    @property
    def regiontree(self):
        if self._regiontree_cached is None:
            self._regiontree_cached = Region(self.name,self,ParcellationIndex(None,None))
            try:
                    self._regiontree_cached.children = tuple( 
                        Region.from_json(regiondef,self) 
                        for regiondef in self._regiondefs )
            except Exception as e:
                logger.error(f"Could not generate child regions for {self.name}")
                raise(e)
        return self._regiontree_cached

    def get_volume_src(self, space: Space):
        """
        Get volumes sources for the parcellation in the requested template space.

        Parameters
        ----------
        space : Space
            template space

        Yields
        ------
        A list of volume sources
        """
        if not self.supports_space(space):
            raise ValueError('Parcellation "{}" does not provide volume sources for space "{}"'.format(
                str(self), str(space) ))
        return self.volume_src[space]

    @cached
    def get_map(self, space=None, maptype:Union[str,MapType]=MapType.LABELLED):
        """
        Get the volumetric maps for the parcellation in the requested
        template space. This might in general include multiple 
        3D volumes. For example, the Julich-Brain atlas provides two separate
        maps, one per hemisphere. Per default, multiple maps are concatenated into a 4D
        array, but you can choose to retrieve a dict of 3D volumes instead using `return_dict=True`.

        Parameters
        ----------
        space : Space or str
            template space specification 
        maptype : MapType (default: MapType.LABELLED)
            Type of map requested (e.g., continous or labelled, see commons.MapType)
            Use MapType.CONTINUOUS to request probability maps.

        Yields
        ------
        A ParcellationMap representing the volumetric map.
        """
        if isinstance(maptype,str):
            maptype = MapType[maptype.upper()]
        if space is None:
            spaceobj = next(iter(self.volume_src.keys()))
            if len(self.volume_src)>1:
                logger.warning(f'Parcellation "{str(self)}" provides maps in multiple spaces, but no space was specified.\nUsing the first, "{str(space)}"')
        else:
            spaceobj = spaces[space]

        if not self.supports_space(spaceobj):
            raise ValueError(f'Parcellation "{self.name}" does not provide a map for space "{spaceobj.name}"')

        return parcellationmap.create_map(self,spaceobj,maptype)

    @property
    def labels(self):
        return self.regiontree.labels

    @property
    def names(self):
        return self.regiontree.names

    def supports_space(self,space : Space):
        """
        Return true if this parcellation supports the given space, else False.
        """
        spaceobj = spaces[space]
        return spaceobj in self.volume_src.keys()

    def decode_region(self,regionspec:Union[str,int,ParcellationIndex,Region]):
        """
        Given a unique specification, return the corresponding region.
        The spec could be a label index, a (possibly incomplete) name, or a
        region object.
        This method is meant to definitely determine a valid region. Therefore, 
        if no match is found, it raises a ValueError. If it finds multiple
        matches, it tries to return only the common parent node. If there are
        multiple remaining parent nodes, which is rare, a custom group region is constructed.

        Parameters
        ----------
        regionspec : any of 
            - a string with a possibly inexact name, which is matched both
              against the name and the identifier key, 
            - an integer, which is interpreted as a labelindex,
            - a region object
            - a full ParcellationIndex

        Return
        ------
        Region object
        """
        candidates = self.regiontree.find(regionspec,select_uppermost=True)
        if not candidates:
            raise ValueError("Regionspec {} could not be decoded under '{}'".format(
                regionspec,self.name))
        elif len(candidates)==1:
            return candidates[0]
        else:
            logger.debug(f"The specification '{regionspec}' resulted more than one region. A group region is returned.")
            return Region._build_grouptree(candidates,self)


    @cached
    def find_regions(self,regionspec):
        """
        Find regions with the given specification in this parcellation.

        Parameters
        ----------
        regionspec : any of 
            - a string with a possibly inexact name, which is matched both
              against the name and the identifier key, 
            - an integer, which is interpreted as a labelindex
            - a region object

        Yield
        -----
        list of matching regions
        """
        return self.regiontree.find(regionspec)

    def __str__(self):
        return self.name

    def __repr__(self):
        return  self.name

    def __eq__(self,other):
        """
        Compare this parcellation with other objects. If other is a string,
        compare to key, name or id.
        """
        if isinstance(other,Parcellation):
            return self.id==other.id
        elif isinstance(other,str):
            return any([
                self.name==other, 
                self.key==other,
                self.id==other])
        else:
            raise ValueError("Cannot compare object of type {} to Parcellation".format(type(other)))

    def __iter__(self):
        """
        Returns an iterator that goes through all regions in this parcellation
        """
        return self.regiontree.__iter__()

    def __getitem__(self,regionspec:Union[str,int]):
        """
        Retrieve a region object from the parcellation by labelindex or partial name.
        """
        return self.decode_region(regionspec)

    @staticmethod
    def from_json(obj):
        """
        Provides an object hook for the json library to construct a Parcellation
        object from a json string.
        """
        required_keys = ['@id','name','shortName','volumeSrc','regions']
        if any([k not in obj for k in required_keys]):
            return obj

        # create the parcellation, it will create a parent region node for the regiontree.
        p = Parcellation(obj['@id'], obj['shortName'], regiondefs=obj['regions'])
        
        if 'volumeSrc' in obj:
            p.volume_src = { spaces[space_id] : {
                key : [
                    volumesrc.from_json(v_src) for v_src in v_srcs
                ] for key, v_srcs in key_vsrcs.items()
            } for space_id, key_vsrcs in obj['volumeSrc'].items() }
        
        if '@version' in obj:
            p.version = ParcellationVersion.from_json(obj['@version'])

        if 'modality' in obj:
            p.modality = obj['modality']

        if 'description' in obj:
            p.description = obj['description']
        if 'publications' in obj:
            p.publications = obj['publications']

        if 'originDatasets' in obj:
            origin_datainfos=[OriginDataInfo.from_json(ds) for ds in obj.get('originDatasets', [])]
            p.origin_datainfos=[f for f in origin_datainfos if f is not None]

        logger.debug(f'Adding parcellation "{str(p)}"')
        return p

REGISTRY = ConfigurationRegistry('parcellations', Parcellation)



# Copyright 2018-2020 Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ctypes import ArgumentError
from . import logger
from . import retrieval
from . import arrays
import numpy as np
import nibabel as nib
from cloudvolume.exceptions import OutOfBoundsError
from cloudvolume import CloudVolume,Bbox
from abc import ABC,abstractmethod
import os
import json

def from_json(obj):
    """
    Provides an object hook for the json library to construct a VolumeSrc
    object from a json stream.
    """
    if obj.get('@type')!="fzj/tmp/volume_type/v0.0.1":
        raise NotImplementedError(f"Cannot build VolumeSrc from this json spec: {obj}")

    volume_type=obj.get('volume_type')
    detail = obj.get('detail')
    url = obj.get('url')

    if volume_type=='nii':
        return NiftiVolume( obj['@id'], obj['name'],url=url,
                detail=detail,zipped_file=obj.get('zipped_file') )

    if volume_type=='neuroglancer/precomputed':
        transform_nm = np.identity(4)
        vsrc_labelindex = None
        if detail is not None and 'neuroglancer/precomputed' in detail:
            if 'transform' in detail['neuroglancer/precomputed']:
                transform_nm = np.array(detail['neuroglancer/precomputed']['transform'])
            if 'labelIndex' in detail['neuroglancer/precomputed']:
                vsrc_labelindex = int(detail['neuroglancer/precomputed']['labelIndex'])
        return NgVolume( obj['@id'], obj['name'],url=url,
                detail=detail, transform_nm=transform_nm)
        #img = volume.build_image(resolution_mm=resolution_mm)
        #if vsrc_labelindex is not None:
        #    # this volume src has a local labelindex - we apply it
        #    img.dataobj[img.dataobj!=vsrc_labelindex] = 0
        #return img

    if volume_type=='detailed maps':
        vsrc = VolumeSrc(obj['@id'], obj['name'],url=None,detail=detail)
        vsrc.volume_type=volume_type
        return vsrc

    # arbitrary volume type

    vsrc = VolumeSrc( obj['@id'], obj['name'],url=url,
            detail=detail)
    vsrc.volume_type = volume_type
    return vsrc

class VolumeSrc:

    def __init__(self, identifier, name, url, detail=None ):
        """
        Construct a new volume source.

        Parameters
        ----------
        identifier : str
            A unique identifier for the source
        name : str
            A human-readable name
        volume_type : str
            Type of volume source, clarifying the data format. Typical names: "nii", "neuroglancer/precomputed".
            TODO create a fixed list (enum?) of supported types, or better: derive types from VolumeSrc object
        url : str
            The URL to the volume src, typically a url to the corresponding image or tilesource.
        detail : dict
            Detailed information. Currently only used to store a transformation matrix  for neuroglancer tilesources.
        zipped_file : str
            The filename to be extracted from a zip file. If given, the url is
            expected to point to a downloadable zip archive. Currently used to
            extreact niftis from zip archives, as for example in case of the
            MNI reference templates.
        """
        self.id = identifier
        self.name = name
        self.url = url
        if 'SIIBRA_URL_MODS' in os.environ and url:
            mods = json.loads(os.environ['SIIBRA_URL_MODS'])
            for old,new in mods.items():
                self.url = self.url.replace(old,new)
            if self.url!=url:
                logger.warning(f'Applied URL modification\nfrom {url}\nto   {self.url}') 
        self.volume_type = None
        self.detail = {} if detail is None else detail
        self.space = None

    def __str__(self):
        return f'{self.volume_type} {self.url}'

    def get_url(self):
        return self.url


class ImageProvider(ABC):

    @abstractmethod
    def fetch(self,resolution_mm=None, voi=None, mapindex=None, clip=False):
        """
        Provide access to image data.
        """
        pass

    @abstractmethod
    def get_shape(self,resolution_mm=None):
        """
        Return the shape of the image volume.
        """
        pass
    
    @abstractmethod
    def is_float(self):
        """
        Return True if the data type of the volume is a float type.
        """
        pass

    def is_4D(self):
        return len(self.get_shape())==4


class NiftiVolume(VolumeSrc,ImageProvider):

    _image_cached=None

    def __init__(self,identifier, name, url, detail=None, zipped_file=None):
        super().__init__(identifier, name, url, detail=detail)
        self.zipped_file = zipped_file
        self.volume_type = 'nii'

    @property
    def image(self):
        """
        We download the nifti file only once it is needed.
        """
        if self._image_cached is None:
            filename = retrieval.download_file(self.url,ziptarget=self.zipped_file)
            if filename is None:
                raise RuntimeError(f"Error retrieving Nifti file from {self.url}")
            self._image_cached=nib.Nifti1Image.from_filename(filename)
            if np.min(self._image_cached.dataobj)<0:
                logger.warn(f'Negative values in {filename} ({str(self)})')
        return self._image_cached

    def fetch(self,resolution_mm=None,voi=None,mapindex=None,clip=False):
        """
        Loads and returns a Nifti1Image object representing the volume source.

        Parameters
        ----------
        resolution_mm : float or None (Default: None)
            Request the template at a particular physical resolution in mm. If None,
            the native resolution is used.
            Currently, this only works for neuroglancer volumes.
        voi : SpaceVOI
            optional volume of interest
        clip : Boolean, default: False
            if True, generates an object where the image data array is cropped to its bounding box (with properly adjusted affine matrix)
        """
        if clip and voi:
            raise ArgumentError("voi and clip cannot only be requested independently")
        shape = self.get_shape()
        img = None
        if resolution_mm is not None:
            raise NotImplementedError(f"NiftiVolume does not support to specify image resolutions (but {resolution_mm} was given)")

        if mapindex is None:
            img = self.image
        elif len(shape)!=4 or mapindex>=shape[3]:
            raise IndexError(f"Mapindex of {mapindex} provided for fetching from NiftiVolume, but its shape is {shape}.")
        else:
            img = nib.Nifti1Image(
                dataobj=self.image.dataobj[:,:,:,mapindex],
                affine=self.image.affine)
        assert(img is not None)

        bb_vox = None
        if voi is not None:
            bb_vox = voi.transform_bbox(np.linalg.inv(img.affine))
        elif clip:
            # determine bounding box by cropping the nonzero values
            bb = arrays.bbox3d(img.dataobj)
            bb_vox = Bbox(bb[:3,0],bb[:3,1])
        
        if bb_vox is not None:
            (x0,y0,z0),(x1,y1,z1) = bb_vox.minpt,bb_vox.maxpt
            shift = np.identity(4)
            shift[:3,-1] = bb_vox.minpt
            img = nib.Nifti1Image(
                dataobj = img.dataobj[x0:x1,y0:y1,z0:z1],
                affine = np.dot(img.affine,shift))

        return img

    def get_shape(self,resolution_mm=None):
        if resolution_mm is not None:
            raise NotImplementedError("NiftiVolume does not support to specify different image resolutions")
        return self.image.shape

    def is_float(self):
        return self.image.dataobj.dtype.kind=='f'


class NgVolume(VolumeSrc,ImageProvider):

    # Gigabyte size that is considered feasible for ad-hoc downloads of
    # neuroglancer volume data. This is used to avoid accidental huge downloads.
    gbyte_feasible = 0.5
    _cached_volume = None
    
    def __init__(self,identifier,name,url,detail=None,transform_nm=np.identity(4)):
        """
        ngsite: base url of neuroglancer http location
        transform_nm: optional transform to be applied after scaling voxels to nm
        """
        super().__init__(identifier,name,url,detail)
        logger.debug(f'Constructing NgVolume "{name}"')
        self.volume_type = "neuroglancer/precomputed"
        self.transform_nm = transform_nm
        r = retrieval.cached_get(url+'/info')
        self.info = json.loads(r.decode())
        self.url = url
        self.nbytes = np.dtype(self.info['data_type']).itemsize
        self.num_scales = len(self.info['scales'])
        self.mip_resolution_mm = { i:np.min(v['resolution'])/(1000**2)
                for i,v in enumerate(self.info['scales']) }
        self.resolutions_available = {
                np.min(v['resolution'])/(1000**2) : {
                    'mip':i,
                    'GBytes':np.prod(v['size'])*self.nbytes/(1024**3)
                    }
                for i,v in enumerate(self.info['scales']) }
        self.helptext = "\n".join(["{:7.4f} mm {:10.4f} GByte".format(k,v['GBytes']) 
            for k,v in self.resolutions_available.items()])

    @property
    def volume(self):
        """
        We implement this as a property so that the CloudVolume constructor is only called lazily.
        """
        if not self._cached_volume:
            self._cached_volume = CloudVolume(self.url,fill_missing=True,progress=False)
        return self._cached_volume
    
    def largest_feasible_resolution(self,voi=None):
        # returns the highest resolution in millimeter that is available and
        # still below the threshold of downloadable volume sizes.
        if voi is None:
            return min([res
                for res,v in self.resolutions_available.items()
                if v['GBytes']<self.gbyte_feasible ])
        else:
            gbytes = {res_mm:voi.transform_bbox(
                        np.linalg.inv(self.build_affine(res_mm))).volume()*self.nbytes/(1024**3)
                        for res_mm in self.resolutions_available.keys() }
            return min([res for res,gb in gbytes.items()
                    if gb<self.gbyte_feasible ])   

    def _resolution_to_mip(self,resolution_mm,voi):
        """
        Given a resolution in millimeter, try to determine the mip that can
        be applied.

        Parameters
        ----------
        resolution_mm : float or None
            Physical resolution in mm.
            If None, the smallest availalbe resolution is used (lowest image size)
            If -1, tha largest feasible resolution is used.
        """
        mip = None
        if resolution_mm is None:
            mip = self.num_scales-1
        elif resolution_mm==-1:
            maxres = self.largest_feasible_resolution(voi=voi)
            mip = self.resolutions_available[maxres]['mip']
            logger.info(f'Using largest feasible resolution of {maxres} mm (mip={mip})')
        elif resolution_mm in self.resolutions_available:
            mip = self.resolutions_available[resolution_mm]['mip']
        if mip is None:
            raise ValueError(f'Requested resolution of {resolution_mm} mm not available.\n{self.helptext}')
        return mip
            
    def build_affine(self,resolution_mm=None,voi=None):
        """
        Builds the affine matrix that maps voxels 
        at the given resolution to physical space.

        Parameters:
        -----------
        resolution_mm : float, or None
            desired resolution in mm. 
            If None, the smallest is used.
            If -1, the largest feasible is used.
        """
        loglevel = logger.getEffectiveLevel()
        logger.setLevel("ERROR")
        mip = self._resolution_to_mip(resolution_mm=resolution_mm,voi=voi)
        effective_res_mm = self.mip_resolution_mm[mip]
        logger.setLevel(loglevel)

        # if a volume of interest is given, apply the offset
        shift = np.identity(4)
        if voi is not None:
            minpt_vox =np.dot(
                np.linalg.inv(self.build_affine(effective_res_mm)),
                np.r_[voi.minpt,1])[:3]
            logger.debug(f"Affine matrix respects volume of interest shift {voi.minpt}")
            shift[:3,-1] = minpt_vox

        # scaling from voxel to nm
        resolution_nm = self.info['scales'][mip]['resolution']
        scaling = np.identity(4)
        for i in range(3):
            scaling[i,i] = resolution_nm[i]

        # optional transform in nanometer space
        affine = np.dot(self.transform_nm,scaling)
            
        # warp from nm to mm   
        affine[:3,:]/=1000000.

        return np.dot(affine,shift)
        #return affine

    def _load_data(self,resolution_mm,voi:Bbox):
        """
        Actually load image data.
        TODO: Check amount of data beforehand and raise an Exception if it is over a reasonable threshold.
        NOTE: this function caches chunks as numpy arrays (*.npy) to the
        CACHEDIR defined in the retrieval module.
        
        Parameters:
        -----------
        resolution_mm : float, or None
            desired resolution in mm. If none, the full resolution is used.
        """
        mip = self._resolution_to_mip(resolution_mm,voi=voi)
        effective_res_mm = self.mip_resolution_mm[mip]
        logger.debug(f"Loading neuroglancer data at a resolution of {effective_res_mm} mm (mip={mip})")
        
        if voi is not None:
            bbox_vox = voi.transform_bbox(
                np.linalg.inv(self.build_affine(effective_res_mm)))
        else:
            bbox_vox = Bbox([0, 0, 0],self.volume.mip_shape(mip))

        # estimate size and check feasibility
        gbytes = bbox_vox.volume()*self.nbytes/(1024**3)
        if gbytes>NgVolume.gbyte_feasible:
            # TODO would better do an estimate of the acutal data size
            logger.error("Data request is too large (would result in an ~{:.2f} GByte download, the limit is {}).".format(gbytes,self.gbyte_feasible))
            print(self.helptext)
            raise NotImplementedError(f"Request of the whole full-resolution volume in one piece is prohibited as of now due to the estimated size of ~{gbytes:.0f} GByte.")

        # ok, retrieve data no now.
        cachefile = retrieval.cachefile("{}{}{}".format(
            self.url, bbox_vox.serialize(), str(mip)).encode('utf8'),suffix='npy')
        if os.path.exists(cachefile):
            logger.debug("NgVolume loads from cache file {cachefile}")
            return np.load(cachefile)
        else:
            try:
                logger.debug(f"NgVolume downloads (mip={mip}, bbox={bbox_vox}")
                data = self.volume.download(bbox=bbox_vox,mip=mip)
                np.save(cachefile,np.array(data))
                return np.array(data)
            except OutOfBoundsError as e:
                logger.error('Bounding box does not match image.')
                print(str(e))
                return None
        
    def fetch(self,resolution_mm=None,voi=None,mapindex=None,clip=False):
        """
        Compute and return a spatial image for the given mip.
        
        Parameters:
        -----------
        resolution_mm : desired resolution in mm
        voi : SpaceVOI
            optional volume of interest
        """
        if clip:
            raise NotImplementedError("Automatic clipping is not yet implemented for neuroglancer volume sources.")
        if mapindex is not None:
            raise NotImplementedError(f"NgVolume does not support access by map index (but {mapindex} was given)")
        return nib.Nifti1Image(
            self._load_data(resolution_mm=resolution_mm,voi=voi),
            affine=self.build_affine(resolution_mm=resolution_mm,voi=voi)
        )

    def get_shape(self,resolution_mm=None):
        mip = self._resolution_to_mip(resolution_mm,voi=None)
        return self.info['scales'][mip]['size']

    def is_float(self):
        return np.dtype(self.info['data_type']).kind=='f'

    def __hash__(self):
        return hash(self.url)+hash(self.transform_nm)
    
    def _enclosing_chunkgrid(self,mip, bbox_phys):
        """
        Produce grid points representing the chunks of the mip 
        which enclose a given bounding box. The bounding box is given in
        physical coordinates, but the grid is returned in voxel spaces of the
        given mip.
        """
        
        # some helperfunctions to produce the smallest range on a grid enclosing another range 
        cfloor = lambda x,s: int(np.floor(x/s)*s)
        cceil = lambda x,s: int(np.ceil(x/s)*s)+1
        crange = lambda x0,x1,s: np.arange(cfloor(x0,s),cceil(x1,s),s)
        
        # project the bounding box to the voxel grid of the selected mip
        bb = np.dot(np.linalg.inv(self.build_affine(self.mip_resolution_mm[mip])),bbox_phys)
        
        # compute the enclosing chunk grid
        chunksizes = self.volume.scales[mip]['chunk_sizes'][0]
        x,y,z = [crange(bb[i][0],bb[i][1],chunksizes[i]) 
                 for i in range(3)]
        xx,yy,zz = np.meshgrid(x,y,z)
        return np.vstack([xx.ravel(),yy.ravel(),zz.ravel(),zz.ravel()*0+1])





# Copyright 2018-2020 Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import PIL.Image as Image
from os import path
from collections import namedtuple
import re

from .feature import RegionalFeature
from .extractor import FeatureExtractor
from ..ebrains import Authentication
from ..termplot import FontStyles as style
from .. import ebrains, retrieval, logger

try:
    import matplotlib.pyplot as plt
    HAVE_PLT=True
except Exception as e:
    HAVE_PLT=False


RECEPTOR_SYMBOLS = {
        "5-HT1A": { 
            "receptor" : {
                "latex" : "$5-HT_{1A}$", 
                "markdown" : "5-HT<sub>1A</sub>", 
                "name" : "5-hydroxytryptamine receptor type 1A" 
                }, 
            'neurotransmitter' : { 
                "label" : "5-HT",
                "latex" : "$5-HT$",
                "markdown" : "5-HT",
                "name" : "serotonin",
                } 
            },
        "5-HT2": {
            "receptor" : {
                "latex" : "$5-HT_2$",
                "markdown" : "5-HT<sub>2</sub>",
                "name" : "5-hydroxytryptamine receptor type 2" 
                },
            'neurotransmitter' : {
                "label" : "5-HT",
                "latex" : "$5-HT$",
                "markdown" : "5-HT",
                "name" : "serotonin" 
                } 
            } ,
        "AMPA": { 
            "receptor" : {
                "latex" : "$AMPA$",
                "markdown" : "AMPA",
                "name" : "alpha-amino-3hydroxy-5-methyl-4isoxazolepropionic acid receptor" 
                },
            'neurotransmitter' : { 
                "label" : "Glu",
                "latex" : "$Glu$",
                "markdown" : "Glu",
                "name" : "glutamate" 
                } 
            },
        "BZ": { 
            "receptor" : {
                "latex" : "$GABA_A/BZ$",
                "markdown" : "GABA<sub>A</sub>/BZ",
                "name" : "gamma-aminobutyric acid receptor type A / benzodiazepine associated binding site" 
                },
            'neurotransmitter' : { 
                "label" : "GABA",
                "latex" : "$GABA$",
                "markdown" : "GABA",
                "name" : "gamma-aminobutyric acid" 
                } 
            },
        "D1": { 
            "receptor" : {
                "latex" : "$D_1$", 
                "markdown" : "D<sub>1</sub>", 
                "name" : "dopamine receptor type 1" 
                } , 
            'neurotransmitter' : { 
                "label" : "DA", 
                "latex" : "$DA$", 
                "markdown" : "DA", 
                "name" : "dopamine" 
                }
            },
            "GABAA": {
                    "receptor" : {
                        "latex" : "$GABA_A$",
                        "markdown" : "GABA<sub>A</sub>",
                        "name" : "gamma-aminobutyric acid receptor type A" 
                        },
                    'neurotransmitter' : {
                        "label" : "GABA",
                        "latex" : "$GABA$",
                        "markdown" : "GABA",
                        "name" : "gamma-aminobutyric acid" 
                        }
                    },
        "GABAB": {
                "receptor" : {
                    "latex" : "$GABA_B$",
                    "markdown" : "GABA<sub>B</sub>",
                    "name" : "gamma-aminobutyric acid receptor type B" 
                    },
                'neurotransmitter' : {
                    "label" : "GABA",
                    "latex" : "$GABA$",
                    "markdown" : "GABA",
                    "name" : "gamma-aminobutyric acid" }
                },
        "M1": {
                "receptor" : {
                    "latex" : "$M_1$",
                    "markdown" : "M<sub>1</sub>",
                    "name" : "muscarinic acetylcholine receptor type 1" 
                    },
                'neurotransmitter' : {
                    "label" : "ACh",
                    "latex" : "$ACh$",
                    "markdown" : "ACh",
                    "name" : "acetylcholine" 
                    }
                },
        "M2": {
                "receptor" : {
                    "latex" : "$M_2$",
                    "markdown" : "M<sub>2</sub>",
                    "name" : "muscarinic acetylcholine receptor type 2" 
                    },
                'neurotransmitter' : {
                    "label" : "ACh",
                    "latex" : "$ACh$",
                    "markdown" : "ACh",
                    "name" : "acetylcholine" 
                    }
                },
        "M3": {
                "receptor" : {
                    "latex" : "$M_3$",
                    "markdown" : "M<sub>3</sub>",
                    "name" : "muscarinic acetylcholine receptor type 3" 
                    },
                'neurotransmitter' : {
                    "label" : "ACh",
                    "latex" : "$ACh$",
                    "markdown" : "ACh",
                    "name" : "acetylcholine" 
                    }
                },
        "NMDA": {
                "receptor" : {
                    "latex" : "$NMDA$",
                    "markdown" : "NMDA",
                    "name" : "N-methyl-D-aspartate receptor" 
                    },
                'neurotransmitter' : {
                    "label" : "Glu",
                    "latex" : "$Glu$",
                    "markdown" : "Glu",
                    "name" : "glutamate" 
                    }
                },
        "alpha1": {
                "receptor" : {
                    "latex" : "$\\alpha_1$",
                    "markdown" : "&#945<sub>1</sub>",
                    "name" : "alpha-1 adrenergic receptor" 
                    },
                'neurotransmitter' : {
                    "label" : "NE",
                    "latex" : "$NE$",
                    "markdown" : "NE",
                    "name" : "norepinephrine" 
                    }
                },
        "alpha2": {
                "receptor" : {
                    "latex" : "$\\alpha_2$",
                    "markdown" : "&#945<sub>2</sub>",
                    "name" : "alpha-2 adrenergic receptor" 
                    },
                'neurotransmitter' : {
                    "label" : "NE",
                    "latex" : "$NE$",
                    "markdown" : "NE",
                    "name" : "norepinephrine" 
                    }
                },
        "alpha4beta2": {
                "receptor" : {
                    "latex" : "$\\alpha_4\\beta_2$",
                    "markdown" : "&#945<sub>4</sub>&#946<sub>2</sub>",
                    "name" : "alpha-4 beta-2 nicotinic receptor" 
                    },
                'neurotransmitter' : {
                    "label" : "ACh",
                    "latex" : "$ACh$",
                    "markdown" : "ACh",
                    "name" : "acetylcholine"
                    }
                },
        "kainate": {
                "receptor" : {
                    "latex" : "$kainate$",
                    "markdown" : "kainate",
                    "name" : "kainate receptors" 
                    },
                'neurotransmitter' : {
                    "label" : "Glu",
                    "latex" : "$Glu$",
                    "markdown" : "Glu",
                    "name" : "glutamate"
                    }
                },
        "mGluR2_3": {
                "receptor" : {
                    "latex" : "$mGluR2/3$",
                    "markdown" : "mGluR2/3",
                    "name" : "metabotropic glutamate receptor type 2 and 3"
                    },
                'neurotransmitter' : {
                    "label" : "Glu",
                    "latex" : "$Glu$",
                    "markdown" : "Glu",
                    "name" : "glutamate"
                    }
                }
        }

def get_bytestream_from_file(url):
    fname = retrieval.download_file(url)
    with open(fname, 'rb') as f:
        return io.BytesIO(f.read())

def unify_stringlist(L: list):
    """ Adds asterisks to strings that appear multiple times, so the resulting
    list has only unique strings but still the same length, order, and meaning. 
    For example: 
        unify_stringlist(['a','a','b','a','c']) -> ['a','a*','b','a**','c']
    """
    assert(all([isinstance(l,str) for l in L]))
    return [L[i]+"*"*L[:i].count(L[i]) for i in range(len(L))]

def edits1(word):
    """
    Produces a list of  all possible edits of a given word that can be produced
    by applying a single character modification.
    From Peter Norvig, see http://norvig.com/spell-correct.html.
    """
    letters    = 'abcdefghijklmnopqrstuvwxyz'
    splits     = [(word[:i], word[i:])    for i in range(len(word) + 1)]
    deletes    = [L + R[1:]               for L, R in splits if R]
    transposes = [L + R[1] + R[0] + R[2:] for L, R in splits if len(R)>1]
    replaces   = [L + c + R[1:]           for L, R in splits if R for c in letters]
    inserts    = [L + c + R               for L, R in splits for c in letters]
    return set(deletes + transposes + replaces + inserts)

def decode_tsv(url):
    bytestream = get_bytestream_from_file(url)
    header = bytestream.readline()
    lines = [l.strip() 
            for l in bytestream.readlines() 
            if len(l.strip())>0]
    sep = b'{' if b'{' in lines[0] else b'\t' 
    keys = unify_stringlist([
        n.decode('utf8').replace('"','').replace("'","").strip() 
        for n in header.split(sep)])
    if any(k.endswith('*') for k in keys):
        logger.debug('Redundant headers: {} in file {}'.format(
            "/".join(keys), url))
    assert(len(keys)==len(set(keys)))
    return  { l.split(sep)[0].decode('utf8') : dict(
        zip(keys, [re.sub(r"\\+",r"\\",v.decode('utf8').strip()) for v in l.split(sep)])) 
        for l in lines }


Density = namedtuple('Density','name, mean, std, unit')

class DensityFingerprint():

    unit = None
    labels = []
    meanvals = []
    stdvals = []
    n = 0

    def __init__(self,datadict):
        """
        Create a DensityFingerprint from a data dictionary coming from a
        receptor fingerprint tsv file.
        """
        units = {list(v.values())[3] for v in datadict.values()}
        assert(len(units)==1)
        self.unit=next(iter(units))
        self.labels=list(datadict.keys())
        try:
            mean=[datadict[l]['density (mean)'] for l in self.labels]
            std=[datadict[l]['density (sd)'] for l in self.labels]
        except KeyError as e: 
            print(str(e))
            logger.error('Could not parse fingerprint from this dictionary')
        self.meanvals=[float(m) if m.isnumeric() else 0 for m in mean]
        self.stdvals=[float(s) if s.isnumeric() else 0 for s in std]

    def __getitem__(self,index):
        assert(index<len(self.labels))
        return Density(
                name=self.labels[index],
                mean=self.meanvals[index],
                std=self.stdvals[index],
                unit=self.unit)

    def __iter__(self):
            self.n = 0
            return self

    def __next__(self):
        if self.n < len(self.labels):
            self.n += 1
            return self[self.n-1]
        else:
            raise StopIteration

    def __str__(self):
        return "\n".join(
                "{d.name:15.15s} {d.mean:8.1f} {d.unit} (+/-{d.std:5.1f})".format(d=d)
                for d in iter(self))


class ReceptorDistribution(RegionalFeature):
    """
    Reprecent a receptor distribution dataset with fingerprint, profiles and
    autoradiograph samples. This implements a lazy loading scheme.
    TODO lazy loading could be more elegant.
    """

    def __init__(self, region, kg_result):

        RegionalFeature.__init__(self,region)
        self.active = False
        self.name = kg_result["name"]
        self.files = kg_result["files"]
        self.info = kg_result['description']
        self.identifier = kg_result['identifier']
        self.url = "https://search.kg.ebrains.eu/instances/Dataset/{}".format(self.identifier)
        self.modality = kg_result['modality']
        self.__profiles = {}
        self.__autoradiographs = {}
        self.__fingerprint = None
        self.__profile_unit = None

    @property
    def autoradiographs(self):
        self._load()
        return self.__autoradiographs

    @property
    def fingerprint(self):
        self._load()
        return self.__fingerprint

    @property
    def profile_unit(self):
        self._load()
        return self.__profile_unit

    @property
    def profiles(self):
        self._load()
        return self.__profiles

    def _load(self):

        if self.active:
            return

        logger.debug('Loading receptor data for'+self.region.name)

        for url in self.files:

            # Receive cortical profiles, if any
            if re.match(r".*_pr[._]",url):
                suffix = path.splitext(url)[-1]
                if suffix == '.tsv':
                    rtype, basename = url.split("/")[-2:]
                    if rtype in basename:
                        data = decode_tsv(url)
                        units = {list(v.values())[3] for v in data.values()}
                        assert(len(units)==1)
                        self.__profile_unit=next(iter(units))
                        # column headers are sometimes messed up, so we fix the 2nd value
                        densities = {int(k):float(list(v.values())[2]) 
                                for k,v in data.items()
                                if k.isnumeric()}
                        rtype = self._check_rtype(rtype)
                        self.__profiles[rtype] = densities

            # Receive autoradiographs, if any
            if re.match(r".*_ar[._]",url):
                rtype, basename = url.split("/")[-2:]
                if rtype in basename:
                    rtype = self._check_rtype(rtype)
                    self.__autoradiographs[rtype] = url

            # receive fingerprint, if any
            if re.match(r".*_fp[._]",url):
                data = decode_tsv(url) 
                self.__fingerprint = DensityFingerprint(data) 

        self.active = True

    def _check_rtype(self,rtype):
        """ 
        Verify that the receptor type name matches the symbol table. 
        Return if ok, fix if a close match is found, raise Excpetion otherwise.
        """
        if rtype in RECEPTOR_SYMBOLS.keys():
            return rtype
        # fix if only different by 1 letter from a close match in the symbol table
        close_matches = list(edits1(rtype).intersection(RECEPTOR_SYMBOLS.keys()))
        if len(close_matches)==1:
            prev,new = rtype, close_matches[0]
            logger.debug("Receptor type identifier '{}' replaced by '{}' for {}".format(
                prev,new, self.region.name))
            return new
        else:
            logger.warning("Receptor type identifier '{}' is not listed in the corresponding symbol table of region {}. Please verify.".format(
                        rtype, self.region.name))
            return rtype

    def __repr__(self):
        self._load()
        return self.__str__()

    def __str__(self):
        """ Outputs a small table of available profiles and autoradiographs. """
        self._load()
        #if len(self.profiles)+len(self.autoradiographs)==0:
                #return style.BOLD+"Receptor density for area {}".format(self.region)+style.END
        return "\n"+"\n".join(
                [style.BOLD+"Receptor densities for area {}".format(self.region.name)+style.END] +
                [style.ITALIC+"{!s:20} {!s:>8} {!s:>15} {!s:>11}".format(
                    'Type','profile','autoradiograph','fingerprint')+style.END] +
                ["{!s:20} {!s:>8} {!s:>15} {!s:>11}".format(
                    rtype,
                    'x'*(rtype in self.profiles),
                    'x'*(rtype in self.autoradiographs),
                    'x'*(rtype in self.fingerprint))
                    for rtype in RECEPTOR_SYMBOLS.keys()
                    if (rtype in self.profiles 
                        or rtype in self.autoradiographs)] )

    def plot(self,title=None):
        if not HAVE_PLT:
            logger.warning('matplotlib.pyplot not available to siibra. plotting disabled.')
            return None

        self._load()
        import numpy as np
        from collections import deque

        # plot profiles and fingerprint
        fig = plt.figure(figsize=(8,3))
        plt.subplot(121)
        for _,profile in self.profiles.items():
            plt.plot(list(profile.keys()),np.fromiter(profile.values(),dtype='d'))
        plt.xlabel('Cortical depth (%)')
        plt.ylabel("Receptor density\n({})".format(self.profile_unit))
        plt.grid(True)
        if title is not None:
            plt.title(title)
        plt.legend(labels=[l for l in self.profiles],
                   loc="center right", prop={'size': 5})

        ax = plt.subplot(122,projection='polar')
        angles = deque(np.linspace(0, 2*np.pi,  len(self.fingerprint.labels)+1)[:-1][::-1])
        angles.rotate(5)
        angles = list(angles)
        means = [d.mean for d in self.fingerprint]
        stds = [d.mean+d.std for d in self.fingerprint]
        plt.plot(angles+[angles[0]],means+[means[0]],'k-',lw=3)
        plt.plot(angles+[angles[0]],stds+[stds[0]],'k',lw=1)
        ax.set_xticks(angles)
        ax.set_xticklabels([l for l in self.fingerprint.labels])
        #ax.set_yticklabels([])
        ax.tick_params(pad=9,labelsize=9)
        ax.tick_params(axis='y',labelsize=6)
        return fig


class ReceptorQuery(FeatureExtractor):

    _FEATURETYPE = ReceptorDistribution
    __features = None

    def __init__(self,atlas):
        # TODO this could probably be a general pattern of the FeatureExtractor Class.

        FeatureExtractor.__init__(self,atlas)
        if self.__class__.__features is None:
            self._load_features()
        for feature in self.__class__.__features:
            self.register(feature)

    def _load_features(self):
        kg_query = ebrains.execute_query_by_id('minds', 'core', 'dataset', 'v1.0.0', 'siibra_receptor_densities')
        self.__class__.__features = []
        for kg_result in kg_query['results']:
            region_names = [e['name'] for e in kg_result['region']]
            for region_name in region_names:
                try:
                    region = self.parcellation.decode_region(region_name)
                    feature = ReceptorDistribution(region,kg_result)
                    self.__class__.__features.append(feature)
                except ValueError as e:
                    # if the region name cannot be decoded, we skip this feature
                    continue


if __name__ == '__main__':

    auth = Authentication.instance()
    auth.set_token('eyJhbGci....')
    extractor = ReceptorQuery()
    print(extractor)



# Copyright 2018-2020 Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .commons import OriginDataInfo
from . import logger
from .retrieval import cached_get
import requests
import json
from os import environ
import re

kg_feature_full_kwargs={
    'org': 'minds',
    'domain': 'core',
    'schema': 'dataset',
    'version': 'v1.0.0'
}

kg_feature_query_kwargs={
    'params': {
        'vocab': 'https://schema.hbp.eu/myQuery/'
    }
}
KG_REGIONAL_FEATURE_FULL_QUERY_NAME='interactiveViewerKgQuery-v1_0'

class EbrainsDataset:
    def __init__(self, id, name, embargo_status):
        
        self.id = id
        self.name = name
        self.embargo_status = embargo_status
        self._detail = None

    @property
    def detail(self):
        if not self._detail:
            self._load()
        return self._detail

    def _load(self):
        if self.id is None:
            raise Exception('id is required')
        match=re.search(r"([a-f0-9-]+)$", self.id)
        if not match:
            raise Exception('id cannot be parsed properly')
        instance_id=match.group(1)
        result=execute_query_by_id(
            query_id=KG_REGIONAL_FEATURE_FULL_QUERY_NAME, 
            instance_id=instance_id,
            msg=f"Retrieving details for '{self.name or self.id}' from EBRAINS...",
            **kg_feature_query_kwargs,**kg_feature_full_kwargs)
        self._detail = result

    def __str__(self):
        return self.name

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, o: object) -> bool:
        # Check type
        if type(o) is not EbrainsDataset and not issubclass(type(o), EbrainsDataset):
            return False
        # Check id
        return self.id == o.id

class EbrainsDatasetOriginDataInfo(EbrainsDataset, OriginDataInfo):

    @property
    def urls(self):
        return [{
            'doi': f,
        }for f in self.detail.get('kgReference', [])]

    @urls.setter
    def urls(self, val):
        pass

    @property
    def description(self):
        return self.detail.get('description')
    
    @description.setter
    def description(self, val):
        pass

    @property
    def name(self):
        # to prevent inf loop, if _detail is undefined, return id
        if self._detail is None:
            return None
        return self.detail.get('name')

    @name.setter
    def name(self, value):
        pass

    def __init__(self, id):
        EbrainsDataset.__init__(self, id=id, name=None, embargo_status=None)

    def __hash__(self):
        return super(EbrainsDataset, self)

    def __str__(self):
        return super(EbrainsDataset, self)

class Authentication(object):
    """
    Implements the authentication to EBRAINS API with an authentication token. Uses a Singleton pattern.
    """
    _instance = None
    _authentication_token = ''

    def __init__(self):
        raise RuntimeError('Call instance() instead')

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls.__new__(cls)
        return cls._instance

    def get_token(self):
        if self._authentication_token == '':
            try:
                self._authentication_token = environ['HBP_AUTH_TOKEN']
            except KeyError:
                logger.warning('An authentication token must be set as an environment variable: HBP_AUTH_TOKEN')
        return self._authentication_token

    def set_token(self, token):
        logger.info('Updating EBRAINS authentication token.')
        self._authentication_token = token


# convenience function that is importet at the package level
def set_token(token):
    auth = Authentication.instance()
    auth.set_token(token)

authentication = Authentication.instance()

def upload_schema(org, domain, schema, version, query_id, file=None, spec=None):
    """
    Upload a query schema to the EBRAINS knowledge graph.

    Parameters
    ----------
    org : str
        organisation
    domain : str
        domain
    schema : str
        schema
    version : str
        version
    query_id : str
        query_id - Used to execute query without specifiying the spec.
    file : str
        path to file of json to be uploaded (overrides spec, if both present)
    spec : dict
        spec to be uploaded (overriden by file, if both present)

    Yields
    ------
    requests.put object
    """
    url = "https://kg.humanbrainproject.eu/query/{}/{}/{}/{}/{}".format(
    org, domain, schema, version, query_id)
    headers={
        'Content-Type':'application/json',
        'Authorization': 'Bearer {}'.format(authentication.get_token())
    }
    if file is not None:
        return requests.put( url, data=open(file, 'r'),
                headers=headers )
    if spec is not None:
        return requests.put( url, json=spec,
                headers=headers )
    raise ValueError('Either file: str or spec: dict is needed for upload_schema method')

def upload_schema_from_file(file, org, domain, schema, version, query_id):
    """
    TODO needs documentation and cleanup
    """
    r = upload_schema(org, domain, schema, version, query_id, file=file)
    if r.status_code == 401:
        logger.error('Invalid authentication in EBRAINS')
    if r.status_code != 200:
        logger.error('Error while uploading EBRAINS Knowledge Graph query.')


def execute_query_by_id(org, domain, schema, version, query_id, instance_id=None, params={}, msg=None):
    """
    TODO needs documentation and cleanup
    """
    url = "https://kg.humanbrainproject.eu/query/{}/{}/{}/{}/{}/instances{}?databaseScope=RELEASED".format(
        org, domain, schema, version, query_id, '/' + instance_id if instance_id is not None else '' )
    if msg is None:
        msg = "No cached data - querying the EBRAINS Knowledge graph..."
    r = cached_get( url, headers={
        'Content-Type':'application/json',
        'Authorization': 'Bearer {}'.format(authentication.get_token())
        }, 
        msg_if_not_cached=msg,
        params=params)
    return json.loads(r)




# Copyright 2018-2020 Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from abc import ABC
from .feature import Feature
from memoization import cached

class FeatureExtractor(ABC):
    """
    An abstract class for data features extractors of a particular type,
    related to an atlas.
    """

    _FEATURETYPE = Feature

    def __init__(self,atlas):
        self.features = []
        self.parcellation = atlas.selected_parcellation
        self.region = atlas.selected_region

    def pick_selection(self,atlas):
        """
        Returns the list of features from this extractor that are associated with
        the selected region of the given atlas object.
        """
        selection = []
        for feature in self.features:
            if feature.matches(atlas):
                selection.append(feature)
        return selection

    def __str__(self):
        return "\n".join([str(f) for f in self.features])

    def register(self,feature):
        assert(isinstance(feature,self._FEATURETYPE))
        self.features.append(feature)


class FeatureExtractorRegistry:

    def __init__(self):
        self._extractors = defaultdict(list)
        self.modalities = {}
        for cls in FeatureExtractor.__subclasses__():
            modality = str(cls._FEATURETYPE).split("'")[1].split('.')[-1]
            self._extractors[modality].append(cls)
            self.modalities[modality] = cls._FEATURETYPE

    def __dir__(self):
        return list(self._extractors.keys())

    def __contains__(self,index):
        return index in self.__dir__()

    def __getattr__(self,name):
        if name in self._extractors.keys():
            return self._extractors[name]
        else:
            raise AttributeError("No such attribute: {}".format(name))
    
    def __getitem__(self,index):
        if index in self._extractors.keys():
            return self._extractors[index]

    def __str__(self):
        return str(self._extractors)



# Copyright 2018-2020 Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from xml.etree import ElementTree
import numpy as np
import json
import siibra
from siibra import retrieval,spaces,logger
from .feature import SpatialFeature
from .extractor import FeatureExtractor
from os import path

class GeneExpression(SpatialFeature):
    """
    A spatial feature type for gene expressions.
    """

    def __init__(self,gene,space,location,expression_levels,z_scores,probe_ids,donor_info,mri_coord=None):
        """
        Construct the spatial feature for gene expressions measured in a sample.

        Parameters:
        -----------
        space : str       
            Name of 3D reference template space in which this feature is defined
        location : tuple of float    
            3D location of the gene expression sample in physical coordinates
            of the given space
        expression_levels : list of float
            expression levels measured in possibly multiple probes of the same sample
        z_scores : list of float
            z scores measured in possibly multiple probes of the same sample
        probe_ids : list of int
            The probe_ids corresponding to each z_score element
        donor_info : dict (keys: age, race, gender, donor, speciment)
            Dictionary of donor attributes
        mri_coord : tuple  (optional)
            coordinates in original mri space
        """
        SpatialFeature.__init__(self,location,space)
        self.expression_levels = expression_levels
        self.z_scores = z_scores
        self.donor_info = donor_info
        self.gene = gene
        self.probe_ids = probe_ids
        self.mri_coord = mri_coord

    def __str__(self):
        return " ".join([
            "At ("+",".join("{:4.0f}".format(v) for v in self.location)+")",
            " ".join(["{:>7.7}:{:7.7}".format(k,str(v)) for k,v in self.donor_info.items()]),
            "Expression: ["+",".join(["%4.1f"%v for v in self.expression_levels])+"]",
            "Z-score: ["+",".join(["%4.1f"%v for v in self.z_scores])+"]"
            ])

class AllenBrainAtlasQuery(FeatureExtractor):
    """
    Interface to Allen Human Brain Atlas microarray data.
    
    This class connects to the web API of the Allen Brain Atlas:
    © 2015 Allen Institute for Brain Science. Allen Brain Atlas API. 
    Available from: brain-map.org/api/index.html
    Any use of the data needs to be in accordance with their terms of use, see
    https://alleninstitute.org/legal/terms-use/

    - We have samples from 6 different human donors. 
    - Each donor corresponds to exactly 1 specimen (tissue used for study)
    - Each sample was subject to multiple (in fact 4) different probes.
    - The probe data structures contain the list of gene expression of a
      particular gene measured in each sample. Therefore the length of the gene
      expression list in a probe coresponds to the number of samples taken in
      the corresponding donor for the given gene.
    """
    _FEATURETYPE = GeneExpression

    ALLEN_ATLAS_NOTIFICATION=\
"""For retrieving microarray data, siibra connects to the web API of
the Allen Brain Atlas (© 2015 Allen Institute for Brain Science), available
from https://brain-map.org/api/index.html. Any use of the microarray data needs
to be in accordance with their terms of use, as specified at
https://alleninstitute.org/legal/terms-use/."""
    _notification_shown=False


    _BASE_URL = "http://api.brain-map.org/api/v2/data"
    _QUERY = {
        "probe" : _BASE_URL+"/query.xml?criteria=model::Probe,rma::criteria,[probe_type$eq'DNA'],products[abbreviation$eq'HumanMA'],gene[acronym$eq{gene}],rma::options[only$eq'probes.id']",
        "specimen" : _BASE_URL+"/Specimen/query.json?criteria=[name$eq'{specimen_id}']&include=alignment3d",
        "microarray" : _BASE_URL+"/query.json?criteria=service::human_microarray_expression[probes$in{probe_ids}][donors$eq{donor_id}]",
        "gene" : _BASE_URL+"/Gene/query.json?criteria=products[abbreviation$eq'HumanMA']&num_rows=all",
        "factors" : _BASE_URL+"/query.json?criteria=model::Donor,rma::criteria,products[id$eq2],rma::include,age,rma::options[only$eq%27donors.id,dono  rs.name,donors.race_only,donors.sex%27]"
        }

    # there is a 1:1 mapping between donors and specimen for the 6 adult human brains
    _DONOR_IDS = [
            '15496', 
            '14380', 
            '15697', 
            '9861', 
            '12876', 
            '10021'] 
    _SPECIMEN_IDS  = [
            'H0351.1015', 
            'H0351.1012', 
            'H0351.1016', 
            'H0351.2001', 
            'H0351.1009', 
            'H0351.2002']

    # load gene names
    genename_file = path.join(path.dirname(siibra.__file__),'features','gene_names.json')
    with open(genename_file,'r') as f:
        GENE_NAMES = json.load(f)

    def __init__(self,atlas,gene):
        """
        Retrieves probes IDs for the given gene, then collects the
        Microarray probes, samples and z-scores for each donor.
        TODO check that this is only called for ICBM space
        """

        FeatureExtractor.__init__(self,atlas)
        self.gene = gene

        if not self.__class__._notification_shown:
            print(self.__class__.ALLEN_ATLAS_NOTIFICATION) 
            self.__class__._notification_shown=True
        logger.info("Retrieving probe ids for gene {}".format(gene))
        url = self._QUERY['probe'].format(gene=gene)
        response = retrieval.cached_get(url)
        root = ElementTree.fromstring(response)
        num_probes = int(root.attrib['total_rows'])
        probe_ids = [int(root[0][i][0].text) for i in range(num_probes)]

        # get specimen information
        self._specimen = {
                spcid:self._retrieve_specimen(spcid) 
                for spcid in self._SPECIMEN_IDS}
        response = json.loads(retrieval.cached_get(self._QUERY['factors']))
        self.factors = {
                item['id']: {
                    'race' : item['race_only'],
                    'gender' : item['sex'],
                    'age' : int(item['age']['days']/365)
                    }
                for item in response['msg'] }

        # get expression levels and z_scores for the gene
        for donor_id in self._DONOR_IDS:
            self._retrieve_microarray(donor_id,probe_ids)


    def _retrieve_specimen(self,specimen_id):
        """
        Retrieves information about a human specimen. 
        """
        url = self._QUERY['specimen'].format(specimen_id=specimen_id)
        response = json.loads(retrieval.cached_get(
            url,msg_if_not_cached="Retrieving specimen information for id {}".format(
                specimen_id)))
        if not response['success']:
            raise Exception('Invalid response when retrieving specimen information: {}'.format( url))
        # we ask for 1 specimen, so list should have length 1
        assert(len(response['msg'])==1)
        specimen = response['msg'][0]
        T = specimen['alignment3d']
        specimen['donor2icbm'] = np.array([
            [T['tvr_00'], T['tvr_01'], T['tvr_02'], T['tvr_09']],
            [T['tvr_03'], T['tvr_04'], T['tvr_05'], T['tvr_10']],
            [T['tvr_06'], T['tvr_07'], T['tvr_08'], T['tvr_11']] ])
        return specimen

    def _retrieve_microarray(self,donor_id, probe_ids):
        """
        Retrieve microarray data for several probes of a given donor, and
        compute the MRI position of the corresponding tissue block in the ICBM
        152 space to generate a SpatialFeature object for each sample.
        """

        # query the microarray data for this donor
        url = self._QUERY['microarray'].format(
                probe_ids=','.join([str(id) for id in probe_ids]),
                donor_id=donor_id)
        response = json.loads(retrieval.cached_get(url))
        if not response['success']:
            raise Exception('Invalid response when retrieving microarray data: {}'.format( url))

        # store probes
        probes,samples = [response['msg'][n] for n in ['probes','samples']]

        # store samples. Convert their MRI coordinates of the samples to ICBM
        # MNI152 space
        for i,sample in enumerate(samples):

            # coordinate conversion to ICBM152 standard space
            donor = {k:sample['donor'][k] for k in ['name','id']}
            icbm_coord = np.matmul(
                    self._specimen[donor['name']]['donor2icbm'],
                    sample['sample']['mri']+[1] ).T

            # Create the spatial feature
            self.register( GeneExpression( 
                self.gene,
                icbm_coord, 
                spaces.MNI152_2009C_NONL_ASYM,
                expression_levels = [float(p['expression_level'][i]) for p in probes],
                z_scores = [float(p['z-score'][i]) for p in probes],
                probe_ids = [p['id'] for p in probes],
                donor_info = {**self.factors[donor['id']], **donor},
                mri_coord = sample['sample']['mri']
                ))

if __name__ == "__main__":

    featureextractor = AllenBrainAtlasQuery('GABARAPL2')



import unittest
from unittest import mock, TestCase
import os
import importlib
import siibra

class TestConfig1(TestCase):

    def test_exported_variable(self):
        self.assertEqual(
            siibra.config.GITLAB_PROJECT_TAG,
            f"siibra-{siibra.__version__}"
        )

    @mock.patch.dict(os.environ, { 'SIIBRA_CONFIG_GITLAB_PROJECT_TAG': 'develop' }, clear=True)
    def test_when_env_set(self):

        importlib.reload(siibra)
        importlib.reload(siibra.config)
        self.assertEqual(
            siibra.config.GITLAB_PROJECT_TAG,
            "develop"
        )

if __name__ == "__main__":
    unittest.main()
